
/*
 * GENERAL FVM routines
 */

#include "equation.h"
#include "quadrature.h"
#include "multi.h"
#include "mlmc.h"

int NXl, NYl, NZl, NXl_FULL, NYl_FULL, NZl_FULL, CURRENT_LEVEL, UPSCALE;
real dt, t, t_halt, halt_times[NSAVES+SAVE0], dx[NDIR];
volume<ConservedVars, 1> U;
volume<ConservedVars, 1> Flux[NDIR];
volume<ConservedVars, 1> Source;
volume<XtraVars, 1> Xtra;

void initial_data (PrimitiveVars &P, int &i, int &j, int &k) {
  PrimitiveVars QP;
  for (int px=0; px<(HAVE_DIR(X) ? QUAD_NUM : 1); px++)
    for (int py=0; py<(HAVE_DIR(Y) ? QUAD_NUM : 1); py++)
      for (int pz=0; pz<(HAVE_DIR(Z) ? QUAD_NUM : 1); pz++) {
        initial_data (QP, (i + NXl*multix + 0.5 + QUAD_POINTS[px]) * dx[X],
                          (j + NYl*multiy + 0.5 + QUAD_POINTS[py]) * dx[Y],
                          (k + NZl*multiz + 0.5 + QUAD_POINTS[pz]) * dx[Z]);
        if (HAVE_DIR(X)) QP *= QUAD_WEIGHTS[px];
        if (HAVE_DIR(Y)) QP *= QUAD_WEIGHTS[py];
        if (HAVE_DIR(Z)) QP *= QUAD_WEIGHTS[pz];
        P += QP;
      }
}

void model_init()
{
  // initialize time invariant variables to increase performance (setting topography, coefficients, etc.)
  time_invariant_vars_init ();
  
  // initial data
  PrimitiveVars P;
  
  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
        Zero(P);
        initial_data (P, i, j, k);
        P *= INITIAL_DATA_SCALING;
        P += INITIAL_DATA_OFFSET;
        INITIAL_DATA_HOOK
        P.toConservedVars (U[i][j][k]);
      }
    }
  }
  
  // store histogram data
  #if HISTOGRAMS
  HIST_DATA [CURRENT_LEVEL] .push_back ( get_histogram_quantity() );
  #endif
}

void init_fvm () {
  t = 0;
  model_init();
  bc_init();
  step_finalize();      // "finalizes" the trivial timestep t=0
}

void run_fvm () {
  while (t < t_halt) {
    
    if (multi_main && isnan(t)) {
      std::cout << " :: ERROR: time is NaN." << std::endl;
      throw 1;
    }
    
    time_step();
    
    #ifndef QUIET
    if (PRINT_t && multi_main) std::cout << " :: time:  t=" << t << " (dt=" << dt << ")" << std::endl;
    #endif
  
    t += dt;
    
    step_finalize();
    POST_STEP_HOOK
  }
}

real calc_time_step()
{
  vector maxspeed;
  maxspeed.zero();
  
  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
        for (int d = 0; d < NDIR; d++) {
          if (!HAVE_DIR(d)) continue;
          maxspeed[d] = max( maxspeed[d], calc_max_speed(U[i][j][k],Xtra[i][j][k],d) );
        }
      }
    }
  }
  return CFL/(maxspeed[X]/dx[X] + maxspeed[Y]/dx[Y] + maxspeed[Z]/dx[Z]);
}

void step_finalize()
{
  #pragma acc parallel loop copy(Xtra,U)
  
  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
        calc_xtra(Xtra[i][j][k], U[i][j][k], i, j, k);
      }
    }
  }
  
  multi_bc();
  
  #if NX != 1
  set_ghost_x();
  #endif
  
  #if NY != 1
  set_ghost_y();
  #endif
  
  #if NZ != 1
  set_ghost_z();
  #endif
}

#if ENTROPIC_RECONSTRUCTION

void ConservedVars_to_ReconstVars () {
  for (int i = -GX; i < NXl+GX; i++)
    for (int j = -GY; j < NYl+GY; j++)
      ConservedVars_to_ReconstVars (U[i][j][0], Xtra[i][j][0]);
}

void ReconstVars_to_ConservedVars () {
  for (int i = -GX; i < NXl+GX; i++)
    for (int j = -GY; j < NYl+GY; j++)
      ReconstVars_to_ConservedVars (U[i][j][0], Xtra[i][j][0]);
}

#endif
