
/*
 * Declarations of two types of FVM routines:
 *  - GENERAL routines for arbitrary system of conservation laws
 *  - EQUATION SPECIFIC (Burgers', advection, MHD, Euler, Shallow Water, Wave, etc.) routines
 */

#ifndef EQUATION_H
#define EQUATION_H

#include "structures.h"
#include "volumes.h"

// === global variables

extern int NXl, NYl, NZl, NXl_FULL, NYl_FULL, NZl_FULL, CURRENT_LEVEL, UPSCALE;
extern real dt, t, t_halt, halt_times[NSAVES+SAVE0], dx[NDIR];
extern volume<ConservedVars, 1> U;
extern volume<ConservedVars, 1> Flux[NDIR];
extern volume<ConservedVars, 1> Source;
extern volume<XtraVars, 1> Xtra;

// === global functions

static int SIZE (int d) { return d==X?NXl:(d==Y?NYl:NZl); }

static int FLUX_NX (int d) { return d==X?NXl+1:NXl; }
static int FLUX_NY (int d) { return d==Y?NYl+1:NYl; }
static int FLUX_NZ (int d) { return d==Z?NZl+1:NZl; }

// === GENERAL FVM routines (defined in equation.cpp and elsewhere)

// compute needed quantities: mesh sizes, initial data, etc.
void model_init();

// use quadrature to compute average of initial data over mesh cell (i,j,k)
void initial_data (PrimitiveVars &P, const int &i, const int &j, const int &k);

// evaluate initial data at domain point (x,y,z)
void initial_data (PrimitiveVars &P, real x, real y, real z);

// initialize FVM
void init_fvm ();

// calculate timestep using CFL number
real calc_time_step();

// time integration routines
void time_step();

// compute XtraVars and set GHOST cells
void step_finalize();

// run FVM
void run_fvm ();

// Godunov-Powell source term
ConservedVars powell_calc (int i, int j, int k);
void powell_source (vector &us, vector &Bs, AllVariables &lV, AllVariables &rV, int i, int j, int k, int d);

// gravity interaction
ConservedVars gravity_calc (int i, int j, int k);

// applies boundary conditions
void set_ghost_x();
void set_ghost_y();
void set_ghost_z();
void bc_init();

// convert volume of ConservedVars to volume of ReconstructionVars
void ConservedVars_to_ReconstVars();

// convert volume of ReconstructionVars to volume of ConservedVars
void ReconstVars_to_ConservedVars(); 

// === EQUATION SPECIFIC FVM for FLUX routines (defined in corresponding flux_*.cpp)

// calculate flux in the direction d
void calc_point_flux (ConservedVars &F, const AllVariables &V, int d);

// calculate maximum wave speed in the direction d
real calc_max_speed (const ConservedVars &CV, const XtraVars &XV, int d);

// calculate XtraVars from ConservedVars
void calc_xtra (XtraVars &XU, const ConservedVars &U, int i, int j, int k);

// calculate AllVariables from PrimitiveVars
void calc_all_vars (AllVariables &V, const PrimitiveVars &P);

// === EQUATION SPECIFIC FVM routines (defined in corresponding equation_*.cpp)

// initialize time invariant variables to increase performance (setting topography, coefficients, etc.)
void time_invariant_vars_init ();

// corrects boundary conditions (if needed) within 'set_ghost_*' routines
void bc_fix (int i, int j, int k, int l, int m, int n);

// get statistical quantity of interest for histograms
#if HISTOGRAMS
real get_histogram_quantity();
#endif

// apply constraints to variables; for instance, pressure should be non-negative, etc.
void apply_constraints (ConservedVars &CV, XtraVars &XV);

// calculate indicator for limiters
real indicator (int i, int j, int k, int d);

// convert ConservedVars to ReconstructionVars
void ConservedVars_to_ReconstVars (ConservedVars &CV, const XtraVars &XV);

// convert ReconstructionVars to ConservedVars
void ReconstVars_to_ConservedVars (ConservedVars &CV, const XtraVars &XV);

// convert ReconstructionVars to ScaledVars
void ReconstVars_to_ScaledVars (ConservedVars &CV, const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d);

// convert ScaledVars to ReconstructionVars
void ScaledVars_to_ReconstVars (ConservedVars &CV, const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d);
  
#endif
