
/* EQUATION SPECIFIC FVM routines
 * Shallow Water equations
 */

///::title:: Shallow Water Equations
///::vars:: GRAVITY=9.812 BED_OFFSET=0.0 BED_SCALING=1.0 CFL=0.45
///::consts:: STANDARD_TOPOGRAPHY_QUAD=1
///::depends:: topography
///::macro:: INITIAL_DATA_HOOK @ P.h -= Xtra[i][j][0].b; if (P.h < 0.0) P.h = 0.0;
///::macro:: SOURCE_HOOK topography.h compute_source(i,j);

#include "equation.h"
#include "solver-util-ec.h"

real calc_max_speed (const ConservedVars &CV, const XtraVars &XV, int d)
{   
    if (CV.h == 0.0)
      return 0; // skips dry cells
    else
      return fabs(XV.u[d]) + sqrt ( GRAVITY * CV.h );
}

void time_invariant_vars_init ()
{
  // initializing bottom topography
  for (int i = -GX; i < NXl+GX; i++)
    for (int j = -GY; j < NYl+GY; j++)
      Xtra[i][j][0].b = b(i,j);
}

void apply_constraints (ConservedVars &CV, XtraVars &XV)
{
  // Constraints for ConservedVars
  if (CV.h < 0.0) CV.h = 0.0;
  // Constraints for XtraVars
  if (XV.E < 0.0) XV.E = 0.0;
}

#if INDICATOR == 1
real indicator(int i, int j, int k, int d)
{
  ConservedVars &Uijk = U[i][j][k];
  return Uijk.h;
}
#endif

void bc_fix (int i, int j, int k, int l, int m, int n) {
  // corrects water level h -- enforces NEUMANN boundary conditions
  U[i][j][k].h -= Xtra[i][j][k].b - Xtra[l][m][n].b;
}

#if ENTROPIC_RECONSTRUCTION

void ConservedVars_to_ReconstVars (ConservedVars &CV, const XtraVars &XV) {
  // converts conserved variables to entropy variables
  CV.h += XV.b;
  CV.h *= GRAVITY;
  CV.h -= sqr(XV.u[X]) / 2.0;
  CV.h -= sqr(XV.u[Y]) / 2.0;
}

void ReconstVars_to_ConservedVars (ConservedVars &CV, const XtraVars &XV) {
  // converts entropy variables to conserved variables
  CV.h += sqr(XV.u[X]) / 2.0;
  CV.h += sqr(XV.u[Y]) / 2.0;
  CV.h /= GRAVITY;
  CV.h -= XV.b;
}

void ReconstVars_to_ScaledVars (ConservedVars &CV, const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d) {
  // converts entropy variables to scaled Roe variables
  // multiplication by R^T
  real RT [3][3];
  compute_R (RT, lCV, rCV, lXV, rXV, d);
  transpose (RT);
  
  real W [3];
  
  W[0] = CV.h;
  W[1] = CV.hu[X];
  W[2] = CV.hu[Y];
  
  zero(CV);
  for (int a=0; a<3; a++) {
    CV.h     += RT[0][a] * W[a];
    CV.hu[X] += RT[1][a] * W[a];
    CV.hu[Y] += RT[2][a] * W[a];
  }
}

void ScaledVars_to_ReconstVars (ConservedVars &CV, const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d) {
  // converts scaled Roe variables to entropy variables
  // multiplication by the inverse of R^T
  real RTinv [3][3];
  compute_Rinv (RTinv, lCV, rCV, lXV, rXV, d);
  transpose (RTinv);
  
  real W [3];
  
  W[0] = CV.h;
  W[1] = CV.hu[X];
  W[2] = CV.hu[Y];
  
  zero(CV);
  for (int a=0; a<3; a++) {
    CV.h     += RTinv[0][a] * W[a];
    CV.hu[X] += RTinv[1][a] * W[a];
    CV.hu[Y] += RTinv[2][a] * W[a];
  }
}

#endif
