
/* EQUATION SPECIFIC FVM routines for FLUX
 * Flux with perturbed permeabilities (Buckley-Leverett)
 */

///::title:: Flux with perturbed permeabilities (Buckley-Leverett)
///::vars:: OIL_VISCOSITY=1.0 WATER_VISCOSITY=1.0
///::consts:: M_UNIFORM=2 M_NORMAL=0 PERM_PERT_OIL=0.3 PERM_PERT_WATER=0.2

#include "equation.h"
#include "chaos.h"

// relative oil permeability
real perm_oil (const ConservedVars &CV, const XtraVars &XV) {
  return pow (fabs(CV.So), 2 + PERM_PERT_OIL*(2*chaos.m_uniform[0] - 1) );
}

// relative water permeability
real perm_water (const ConservedVars &CV, const XtraVars &XV) {
  return pow (fabs(XV.Sw), 2 + PERM_PERT_WATER*(2*chaos.m_uniform[1] - 1) );
}

// derivative of relative oil permeability
real d_dSo_perm_oil (const ConservedVars &CV, const XtraVars &XV) {
  return ( 2 + PERM_PERT_OIL*(2*chaos.m_uniform[0] - 1) ) * pow (fabs(CV.So), 1 + PERM_PERT_OIL*(2*chaos.m_uniform[0] - 1) );
  // take into account that absolute value is used (different derivative?)
}

// derivative water relative oil permeability
real d_dSo_perm_water (const ConservedVars &CV, const XtraVars &XV) {
  return -( 2 + PERM_PERT_WATER*(2*chaos.m_uniform[1] - 1) ) * pow (fabs(XV.Sw), 1 + PERM_PERT_WATER*(2*chaos.m_uniform[1] - 1) );
}

real lambda_o, lambda_w, d_dSo_lambda_o, d_dSo_lambda_w, VISCOSITY_FRAC;

real calc_max_speed (const ConservedVars &CV, const XtraVars &XV, int d)
{
  real buffer = 0;
  
  VISCOSITY_FRAC = OIL_VISCOSITY / WATER_VISCOSITY;
  
  lambda_o = perm_oil   (CV, XV);
  lambda_w = perm_water (CV, XV);
  
  d_dSo_lambda_o = d_dSo_perm_oil   (CV, XV);
  d_dSo_lambda_w = d_dSo_perm_water (CV, XV);
  
  buffer += d_dSo_lambda_o * (lambda_o + VISCOSITY_FRAC * lambda_w);
  buffer -= (d_dSo_lambda_o + VISCOSITY_FRAC * d_dSo_lambda_w) * lambda_o;
  buffer /= sqr(lambda_o + VISCOSITY_FRAC * lambda_w);
  buffer *= TOTAL_FLOW_RATE * ROCK_PERMEABILITY;
  
  return fabs(buffer);
}

void calc_point_flux(ConservedVars &F, const AllVariables &V, int d)
{
  VISCOSITY_FRAC = OIL_VISCOSITY / WATER_VISCOSITY;
  F.So = perm_oil (V.U, V.XU) / ( perm_oil (V.U, V.XU) + VISCOSITY_FRAC * perm_water (V.U, V.XU) );
  F.So *= TOTAL_FLOW_RATE * ROCK_PERMEABILITY;
}

void calc_xtra(XtraVars &Xijk, const ConservedVars &Uijk, int i, int j, int k)
{
  Xijk.Sw = 1.0 - Uijk.So;
  // Constraints for XtraVars
  if (Xijk.Sw < 0.0) Xijk.Sw = 0.0;
  if (Xijk.Sw > 1.0) Xijk.Sw = 1.0;
}

void calc_all_vars(AllVariables &V, const PrimitiveVars &P)
{
  P.toConservedVars(V.U);
  V.XU.Sw = 1.0 - P.So;
  apply_constraints (V.U, V.XU);
}
