
#include "material-fft.h"

#include "equation.h"
#include "chaos.h"

int NXl_kernel, NYl_kernel, NZl_kernel;
volume <real, 0> c_buffer;

// FFT of the generator of the circulant covariance matrix
fftw_complex* C_FFT;

// FFT of the Gaussian random vector
fftw_complex* chaos_FFT;

// sampling buffer
fftw_complex* buffer;

// prints out complex array
void print_buffer_complex (fftw_complex *buffer, char name) {
  
  if (ACTIVE_DIRS == 3) {
    std::cout << "print_buffer_complex() is not implemented for 3d" << std::endl;
    return;
  }
  
  std::cout << name << ": " << std::endl;
  for (int i=0; i<NXl_kernel; i++) {
    for (int j=0; j<NYl_kernel; j++)
      std::cout << buffer[getLoc(i,j,0)][0] << "+" << buffer[getLoc(i,j,0)][1] << "i ";
    std::cout << std::endl;
  }
}

void material_fft_init () {
  
  // volumes for kernels
  volume <real, 0> *kernel;
  volume <real, 0> *kernel_L = new volume <real, 0> (NX, NY, NZ);
  
  // initializes generator of the circulant covariance matrix (at level l=L)
  vector x1, x2;
  x1.zero(); x2.zero();
  for (int i=0; i<NX; i++) {
    for (int j=0; j<NY; j++)
      for (int k=0; k<NZ; k++) {
        if (HAVE_DIR(X)) x2[X] = 1.0*i*MAXX/NX;
        if (HAVE_DIR(Y)) x2[Y] = 1.0*j*MAXY/NY;
        if (HAVE_DIR(Z)) x2[Z] = 1.0*k*MAXZ/NZ;
        (*kernel_L)[i][j][k] = material_covariance_kernel ( x1, x2 );
      }
  }
  
  #if DEBUG
  std::cout << "L: " << std::endl;
  kernel_L->print();
  #endif
  
  // computes kernel by upscaling kernel_L to level CURRENT_LEVEL (CURRENT_LEVEL+1 if UPSCALE=1)
  kernel = material_fft_upscale_kernel (kernel_L, L-(CURRENT_LEVEL+UPSCALE));
  
  // storing resulting kernel size
  NXl_kernel = kernel->NXl;
  NYl_kernel = kernel->NYl;
  NZl_kernel = kernel->NZl;
  
  #if DEBUG
  std::cout << "k: " << std::endl;
  kernel->print();
  #endif
  
  // allocating memory for arrays
  C_FFT     = (fftw_complex*) fftw_malloc(NXl_kernel * NYl_kernel * NZl_kernel * sizeof(fftw_complex));
  chaos_FFT = (fftw_complex*) fftw_malloc(NXl_kernel * NYl_kernel * NZl_kernel * sizeof(fftw_complex));
  buffer    = (fftw_complex*) fftw_malloc(NXl_kernel * NYl_kernel * NZl_kernel * sizeof(fftw_complex));
  
  if (C_FFT == NULL || chaos_FFT == NULL || buffer == NULL) {
      std::cout << " :: ERROR: out of memory for FFT buffers." << std::endl;
      throw 1;
  }
  
  c_buffer.reinit(NXl_kernel, NYl_kernel, NZl_kernel);
  
  // computes FFT of kernel
  material_fft_r2c (kernel->data, C_FFT);
  
  #if DEBUG
  print_buffer_complex (C_FFT,'C');
  #endif
  
  // make sure imaginary parts are zero
  for (int i=0; i<NXl_kernel/2+1; i++)
    for (int j=0; j<NZl_kernel; j++)
      for (int k=0; k<NZl_kernel; k++) {
        if (C_FFT[getLoc(i,j,k)][1] < 1e-10) {
          C_FFT[getLoc(i,j,k)][1] = 0.0;
        } else {
          std::cout << " :: ERROR: C_FFT has non-zero imaginary parts at index ";
          std::cout << "(" << i << "," << j << "," << k << ")" << std::endl;
          throw 1;
        }
      }
  
  // take a (complex) square root of the C_FFT vector (Cholesky factor)
  for (int i=0; i<NXl_kernel/2+1; i++)
    for (int j=0; j<NYl_kernel; j++)
      for (int k=0; k<NZl_kernel; k++) {
        if (C_FFT[getLoc(i,j,k)][0] >= 0.0)
          C_FFT[getLoc(i,j,k)][0] = sqrt(C_FFT[getLoc(i,j,k)][0]);
        else {
          C_FFT[getLoc(i,j,k)][1] = sqrt(fabs(C_FFT[getLoc(i,j,k)][0]));
          C_FFT[getLoc(i,j,k)][0] = 0.0;
        }
      }
  
  #if DEBUG
  print_buffer_complex (C_FFT,'F');
  #endif
}

void material_fft_draw_sample () {
  
  // compute FFT of the Gaussian random variable vector
  material_fft_r2c (chaos.m_normal, chaos_FFT);
  
  // generate correlated material coefficient in Fourier domain
  for (int i=0; i<NXl_kernel; i++)
    for (int j=0; j<NYl_kernel; j++)
      for (int k=0; k<NZl_kernel; k++) {
        buffer[getLoc(i,j,k)][0] = C_FFT[getLoc(i,j,k)][0] * chaos_FFT[getLoc(i,j,k)][0] - C_FFT[getLoc(i,j,k)][1] * chaos_FFT[getLoc(i,j,k)][1];
        buffer[getLoc(i,j,k)][1] = C_FFT[getLoc(i,j,k)][0] * chaos_FFT[getLoc(i,j,k)][1] + C_FFT[getLoc(i,j,k)][1] * chaos_FFT[getLoc(i,j,k)][0];
      }
  
  // compute inverse FFT
  material_fft_c2r (buffer, c_buffer.data);
  
  #if DEBUG
  std::cout << "c: " << std::endl;
  c_buffer.print();
  #endif
  
  // complete inverse FFT (division by NXl_kernel) and compute exponentials of elements in c_buffer
  c_buffer /= NXl_kernel * NYl_kernel * NZl_kernel;
  c_buffer.exp();
  
  #if DEBUG
  std::cout << "e: " << std::endl;
  c_buffer.print();
  #endif
  
  // computes cell averages needed for coupling the MC estimates of two consecutive levels
  if (UPSCALE && CURRENT_LEVEL+1 <= L)
    material_fft_average_volume (c_buffer);
  
  #if DEBUG
  if (UPSCALE && CURRENT_LEVEL+1 <= L) {
    std::cout << "a: (only first half relevant)" << std::endl;
    c_buffer.print();
  }
  #endif
  
  // SCALING
  if (MATERIAL_SCALING != 1.0)
    c_buffer *= MATERIAL_SCALING;
  
  // OFFSET
  if (MATERIAL_OFFSET != 0.0)
    c_buffer += MATERIAL_OFFSET;
}

void material_fft_exit() {
  fftw_free(C_FFT);
  fftw_free(chaos_FFT);
  fftw_free(buffer);
}

void material_fft_r2c (real *in, fftw_complex *out) {
  
  fftw_plan p_r2c;
  
  int N_kernel [3] = {NXl_kernel, NYl_kernel, NZl_kernel};
  
  p_r2c = fftw_plan_dft_r2c (ACTIVE_DIRS, N_kernel, in, out, FFTW_ESTIMATE);
  
  fftw_execute(p_r2c);
  fftw_destroy_plan(p_r2c);
}

void material_fft_c2r (fftw_complex *in, real *out) {
  
  fftw_plan p_c2r;
  
  int N_kernel [3] = {NXl_kernel, NYl_kernel, NZl_kernel};
  
  p_c2r = fftw_plan_dft_c2r (ACTIVE_DIRS, N_kernel, in, out, FFTW_ESTIMATE);
  
  fftw_execute(p_c2r);
  fftw_destroy_plan(p_c2r);
}

volume <real, 0> * material_fft_upscale_kernel (volume <real, 0> *in, int levels) {
  
  if (levels == 0) {
    // no more upscaling is needed, simply copy (to prevent out == in when levels = 0)
    volume <real, 0> *out = new volume <real, 0> (*in);
    delete in;
    return out;
  }
  
  else {
    
    // initializes volume on a coarser level
    volume <real, 0> *out = new volume <real, 0> (max(in->NXl/2,1), max(in->NYl/2,1), max(in->NZl/2,1));
    
    // shifts for upscaling of the covariance kernel
    int shifts [4] = {-1, 0, 0, 1};
    
    for (int i=0; i<out->NXl; i++)
      for (int j=0; j<out->NYl; j++)
        for (int k=0; k<out->NZl; k++)
          for (int l=0; l<4; l++)
            for (int m=(HAVE_DIR(Y)?0:1); m<(HAVE_DIR(Y)?4:2); m++)
              for (int n=(HAVE_DIR(Z)?0:1); n<(HAVE_DIR(Z)?4:2); n++)
                (*out)[i][j][k] += (*in)[mod(2*i+shifts[l], 2*NXl)][mod(2*j+shifts[m], 2*NYl)][mod(2*k+shifts[n], 2*NZl)];
    
    // scales values according to the number of dimensions
    if (HAVE_DIR(Z)) (*out) /= 64;
    else if (HAVE_DIR(Y)) (*out) /= 16;
    else (*out) /= 4;
    
    delete in;
    return material_fft_upscale_kernel (out, levels-1);
  }
}

void material_fft_average_volume (volume <real, 0> &v) {
  
  // computes cell averages
  for (int i=0; i<NXl_FULL; i++)
    for (int j=0; j<NYl_FULL; j++)
      for (int k=0; k<NZl_FULL; k++) {
        v[i][j][k]    = v[2*i][2*j  ][2*k]   + v[2*i+1][2*j  ][2*k];
        if (HAVE_DIR(Y))
          v[i][j][k] += v[2*i][2*j+1][2*k]   + v[2*i+1][2*j+1][2*k];
        if (HAVE_DIR(Z)) {
          v[i][j][k] += v[2*i][2*j  ][2*k+1] + v[2*i+1][2*j  ][2*k+1];
          v[i][j][k] += v[2*i][2*j+1][2*k+1] + v[2*i+1][2*j+1][2*k+1];
        }
      }
  
  // scales valus according to the number of dimensions
  if (HAVE_DIR(Z)) v /= 8;
  else if (HAVE_DIR(Y)) v /= 4;
  else v /= 2;
}

int getLoc (int i, int j, int k) {
  return i*NYl_kernel*NZl_kernel + j*NZl_kernel + k;
}
