
#include <cstdlib>
#include <iostream>

#include "mlmc.h"
#include "rng.h"
#include "chaos.h"
#include "multi.h"
#include "stats.h"
#include "saver.h"

// === MLMC

int N[NDIR][L+1];
long_int NM[L+1];

// container for random structures
Chaos chaos;

// === volumes for statistical information and performance optimization

volume <ConservedVars, 1> mean_U [NSAVES+SAVE0];
volume <XtraVars, 1>      mean_X [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
volume <ConservedVars, 1> var_U  [NSAVES+SAVE0];
volume <XtraVars, 1>      var_X  [NSAVES+SAVE0];
#endif

volume <ConservedVars, 1> mean_U_l [NSAVES+SAVE0];
volume <XtraVars, 1>      mean_X_l [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
volume <ConservedVars, 1> var_U_l  [NSAVES+SAVE0];
volume <XtraVars, 1>      var_X_l  [NSAVES+SAVE0];
#endif

#if DOWNSAMPLE
volume <ConservedVars, 1> mean_U_l_downsampled [NSAVES+SAVE0];
volume <XtraVars, 1>      mean_X_l_downsampled [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
volume <ConservedVars, 1> var_U_l_downsampled  [NSAVES+SAVE0];
volume <XtraVars, 1>      var_X_l_downsampled  [NSAVES+SAVE0];
#endif
#endif

#if COMPUTE_VARIANCE
volume <ConservedVars, 1> temp_U_l;
volume <XtraVars,      1> temp_X_l;
#endif

#if HISTOGRAMS
std::vector<real> HIST_DATA [L+1];
#endif

// === MLMC routines

// computes MULTI[:][:], NM[:], SAMPLERS[:], N[:][:], MULTI_TOTAL, LOWEST_MULTI_LEVEL
void mlmc_init () {
  
  // !!! IF MODIFYING, also CHECK make.py::MakeMLMC::get_multi_total() for possible necessary modifications
  
  // === computes level resolutions and parallelisation parameters | N[:][:], NM[:], MULTI[:][:]
  
  MULTI[X][L] = MULTIX;
  MULTI[Y][L] = MULTIY;
  MULTI[Z][L] = MULTIZ;
  MULTI[M][L] = MULTIM;
  
  NM[L] = ML;
  // In the case ML = -1 use a file to read the number of samples on every level
  std::ifstream is("Ml.txt");
  if (ML == -1 && !is) 
  {
    std::cout << "In the run folder there is no Ml.txt file!";
    multi_finalise();
    std::exit(1);
  }

  if (ML == -1)
  {
    long int al;
    is >> al;
    //std::cout << "  ML " << al; 
    NM[L] = al;
  }
  
  N[X][L] = NX / MULTI[X][L];
  N[Y][L] = NY / MULTI[Y][L];
  N[Z][L] = NZ / MULTI[Z][L];
  
  for (int l=L-1; l>=0; l--) {
    
    // In the case ML = -1 use a file to read the number of samples on every level
    if (ML == -1) {
      long_int al;
      is >> al;
      NM[l] = al;
    }
    else {
      if (FACTOR * NM[l+1] <= NM_MAX)
        NM[l]  = FACTOR * NM[l+1];
      else {
        std::cerr << " :: ERROR: NM[l] exceeded maximum value of " << NM_MAX << ". Change typedef of 'long_int' in 'numtypes.h'." << std::endl;
        throw 1;
      }
    }
    
    // Maximal cummulative factor allowed in: MULTI[X-Z][l] = MULTI[X-Z][l+1] / 2.
    // Ensures that: MULTI[M][l] <= NM[l+1] for all l in {L-1, ..., 0}.
    int MULTI_FACTOR_MAX = WORK_FACTOR * NM[l+1] / MULTI[M][l+1];
    
    // if DOWNSAMPLE=1, then no recomputation is present; hence MULTI[M][l+1] is no longer of high importance for level = l.
    for (int d=0; d<NDIR; d++) {
      if ( MULTI[d][l+1] != 1 && (MULTI_FACTOR_MAX > 1 || DOWNSAMPLE) ) {
        MULTI_FACTOR_MAX /= 2;
        MULTI[d][l] = MULTI[d][l+1] / 2;
      } else {
        MULTI[d][l] = MULTI[d][l+1];
      }
    }
    
    int MULTI_FACTOR  = (MULTI[X][l+1] / MULTI[X][l]) * (MULTI[Y][l+1] / MULTI[Y][l]) * (MULTI[Z][l+1] / MULTI[Z][l]);
    
    if ( (MULTI[M][l+1] * MULTI_FACTOR) >= WORK_FACTOR )
      MULTI[M][l] = (MULTI[M][l+1] * MULTI_FACTOR) / WORK_FACTOR;
    else
      MULTI[M][l] = 1;
    
    N[X][l] = N[X][l+1] / 2 * (MULTI[X][l+1] / MULTI[X][l]);
    
    if ( NY > 1 )
      N[Y][l] = N[Y][l+1] / 2 * (MULTI[Y][l+1] / MULTI[Y][l]);
    else
      N[Y][l] = N[Y][l+1];
    
    if ( NZ > 1 )
      N[Z][l] = N[Z][l+1] / 2 * (MULTI[Z][l+1] / MULTI[Z][l]);
    else
      N[Z][l] = N[Z][l+1];
    
    if (N[X][l] == 0 || N[Y][l] == 0 || N[Z][l] == 0) {
      std::cerr << " :: ERROR: Too many levels, no further coarsening possible" << std::endl;
      throw 1;
    }
  }
  
  // === computes how many levels should be parallelised | LOWEST_MULTI_LEVEL
  
  LOWEST_MULTI_LEVEL = 0;
  
  // if all levels do the same amount of work, all of them should be parallelised
  if (WORK_FACTOR != 1) {
    // otherwise find the level from which on parallelisation is no longer needed
    for (int l=L-1; l>=0; l--) {
      if (MULTI[M][l] == 1 && MULTI[M][l+1] == 1
          && MULTI[X][l]   == 1 && MULTI[Y][l]   == 1 && MULTI[Z][l]   == 1
          && MULTI[X][l+1] == 1 && MULTI[Y][l+1] == 1 && MULTI[Z][l+1] == 1)
      {
        LOWEST_MULTI_LEVEL = l;
        break;
      }
    }
  }
  
  // === computes needed number of nodes | MULTI_TOTAL
  
  MULTI_TOTAL = 0;
  for (int l=L; l>=LOWEST_MULTI_LEVEL; l--) {
    MULTI_TOTAL += MULTI[M][l] * MULTI[X][l] * MULTI[Y][l] * MULTI[Z][l];
  }
}

void run_compatibility_checks () {
  
  // check for condition MULTIM <= ML (ML=-1 indecates that ML is read from file)
  if (ML !=-1 && MULTIM > ML && multi_main) {
    std::cout << " :: ERROR: MULTIM <= ML is violated: " << MULTIM << " > " << ML << std::endl;
    multi_finalise();
    std::exit(1);
  }
  
  // DOWNSAMPLE=1 and LOAD_BALANCING_TYPE=2 are incompatible! (SAVED samples need to be TRANSFERRED between cores)
  if (DOWNSAMPLE == 1 && LOAD_BALANCING_TYPE == 2  && multi_main) {
    std::cout << " :: ERROR: DOWNSAMPLE=1 and LOAD_BALANCING_TYPE=2 are incompatible!" << std::endl;
    multi_finalise();
    std::exit(1);
  }
}

void display_info () {
  #ifndef QUIET
  if (multi_main) {
    if (L == 0) {
      std::cout << " :: MC-FVM simulation" << std::endl;
      std::cout << " :: INFO: number of cells:   " << NX << "x" << NY << "x" << NZ << std::endl;
      std::cout << " :: INFO: number of samples: " << NM[0] << std::endl;
    }
    else {
      std::cout << " :: MLMC-FVM simulation with " << L+1 << " levels" << std::endl;
      std::cout << " :: INFO: number of cells   (finest mesh level)   : " << N[X][L] * MULTI[X][L] << "x" << N[Y][L] * MULTI[Y][L] << "x" << N[Z][L] * MULTI[Z][L] << std::endl;
      std::cout << " :: INFO: number of samples (finest mesh level)   : " << NM[L] << std::endl;
      std::cout << " :: INFO: number of cells   (coarsest mesh level) : " << (1 << L0X) << "x" << (1 << L0Y) << "x" << (1 << L0Z) << std::endl;
      std::cout << " :: INFO: number of samples (coarsest mesh level) : " << NM[0] << std::endl;
      std::cout << " :: INFO: number of samples (total)               : " << NMT << std::endl;
    }
  }
  //std::cout << std::endl;
  #endif
}

void volumes_init () {
  
  int l = LEVEL_END;
  
  for(int s = 0; s<NSAVES+SAVE0; s++) {
    
    // resizing array to match size of the finest level computed on this particular node
    mean_U [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    mean_X [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    #if COMPUTE_VARIANCE
    var_U  [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    var_X  [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    #endif
    
    // for stabilization of ML-part with exp-log
    #if NUMERICAL_STABILIZATION
      mean_U [s] .exp();
      mean_X [s] .exp();
      #if COMPUTE_VARIANCE
      var_U  [s] .exp();
      var_X  [s] .exp();
      #endif
    #endif
  }
}

// setup meshes for level l
void level_init (int l) {
  
  CURRENT_LEVEL = l;
  
  NXl = N[X][l];
  NYl = N[Y][l];
  NZl = N[Z][l];
  
  // setup cores for domain decomposition and samplers
  MULTIXl = MULTI[X][l];
  MULTIYl = MULTI[Y][l];
  MULTIZl = MULTI[Z][l];
  MULTIMl = MULTI[M][l];
  
  MULTINl = MULTIXl * MULTIYl * MULTIZl;
  
  NXl_FULL = NX > 1 ? NXl * MULTIXl : 1;
  NYl_FULL = NY > 1 ? NYl * MULTIYl : 1;
  NZl_FULL = NZ > 1 ? NZl * MULTIZl : 1;
  
  // mesh parameters
  dx[X] = 1.0*MAXX/NXl_FULL;
  dx[Y] = 1.0*MAXY/NYl_FULL;
  dx[Z] = 1.0*MAXZ/NZl_FULL;
  
  // reinit volumes for resolutions | NXl, NYl, NZl
  U      .reinit (NXl, NYl, NZl);
  Xtra   .reinit (NXl, NYl, NZl);
  Source .reinit (NXl, NYl, NZl);
  for (int i=0; i<NDIR; i++)
    Flux[i].reinit (NXl, NYl, NZl);
}

void display_level_info(int l, char action) {
  #ifndef QUIET
  if (multi_level_root && L != 0) {
    std::cout << " :: LEVEL " << l;
    if (l != 0) std::cout << action;
      std::cout << ": ";
    std::cout << "resolution: " << NXl*MULTIXl << "x" << NYl*MULTIYl << "x" << NZl*MULTIZl;
    std::cout << ", samples: " << NM[l];
    std::cout << ", cores: (" << MULTIXl << "x" << MULTIYl << "x" << MULTIZl << ")x" << MULTIMl;
    std::cout << "=" << MULTIXl*MULTIYl*MULTIZl*MULTIMl << "...";
    std::cout << std::endl;
  }
  #endif
}

void display_level_done(int l, char action) {
  #ifndef QUIET
  if (multi_level_root)
    if (L == 0) std::cout << "  : -> done." << std::endl;
    else {
      std::cout << " :: LEVEL " << l;
      if (l != 0) std::cout << action;
      std::cout << ": -> done." << std::endl;
    }
  #endif
}

unsigned int getSeed (unsigned int level) {
  
  // Setting RUN_NUMBER = -1 will force the programm to read the real RUN_NUMBER from the 
  // file "RUN_NUMBER.txt". This allows to make several runs without recompiling the programm.
  long_int run_number = RUN_NUMBER;
  std::ifstream is("RUN_NUMBER.txt");
  if (RUN_NUMBER == -1 && !is) 
  {
    std::cout << "In the run folder there is no RUN_NUMBER.txt file!";
  }
  if (RUN_NUMBER == -1)
  {
    is >> run_number;
    //std::cout << "  run_number " << run_number; 
  }
  
  unsigned int pos = PRIME_SHIFT + (run_number-1)*L + level;
  
  if (pos >= NPRIMES) {
    if (multi_main) {
      std::cerr << " :: ERROR: Prime number array 'primes' in primes.h is too small!" << std::endl;
      std::cerr << "  -> Decrease PRIME_SHIFT or L or RUN_NUMBER or increace the 'primes' array." << std::endl;
    }
    multi_finalise();
    std::exit(1);
  }
  
  return primes[pos];
}

void rng_and_chaos_init (int l_samples, int l_seed ) {
  rng_init ( getSeed(l_seed) );
  chaos_init ( SAMPLES_OFFSET + multi_rank_samplers * (NM[l_samples]/MULTIMl) );
}

void run_samples (int l, char action, bool compute) {
  
  // initializing level l volumes for statistical information
  for(int s = 0; s<NSAVES+SAVE0; s++) {
    mean_U_l [s] .reinit (NXl, NYl, NZl);
    mean_X_l [s] .reinit (NXl, NYl, NZl);
    #if COMPUTE_VARIANCE
    var_U_l  [s] .reinit (NXl, NYl, NZl);
    var_X_l  [s] .reinit (NXl, NYl, NZl);
    #endif
  }
  
  #if COMPUTE_VARIANCE
  // partition size for MPI calls
  MULTIM_PARTITION_SIZE = NM[l]/MULTIMl;
  
  // temporary volumes for computational efficiency
  temp_U_l .reinit (NXl, NYl, NZl);
  temp_X_l .reinit (NXl, NYl, NZl);
  #endif
  
  #if COMPUTE_KURTOSIS
  volume <ConservedVars, 1> M2 [NSAVES+SAVE0];
  volume <ConservedVars, 1> M3 [NSAVES+SAVE0];
  volume <ConservedVars, 1> M4 [NSAVES+SAVE0];
  
  // loop over all time steps
  for (int s=0; s<NSAVES+SAVE0; s++) {
    M2[s] .reinit (NXl, NYl, NZl);
    M3[s] .reinit (NXl, NYl, NZl);
    M4[s] .reinit (NXl, NYl, NZl);
  }
  
  volume <ConservedVars, 1> delta (NXl, NYl, NZl);
  volume <ConservedVars, 1> delta_n (NXl, NYl, NZl);
  volume <ConservedVars, 1> delta_n2 (NXl, NYl, NZl);
  volume <ConservedVars, 1> term1 (NXl, NYl, NZl);
  #endif
  
  // compute samples
  if (compute) {
    
    #if DOWNSAMPLE
    if (l != L) {
      // loop over all time steps
      for (int s=0; s<NSAVES+SAVE0; s++) {
        mean_U_l_downsampled [s] .reinit (NXl, NYl, NZl);
        mean_X_l_downsampled [s] .reinit (NXl, NYl, NZl);
        #if COMPUTE_VARIANCE
        var_U_l_downsampled  [s] .reinit (NXl, NYl, NZl);
        var_X_l_downsampled  [s] .reinit (NXl, NYl, NZl);
        #endif
      }
      
      // samplers that work on 'downsampled' set of samples
      ACTIVE_MULTIMl = NM[l+1] / (NM[l]/MULTI[M][l]);
    }
    #endif
    
    // loop over all MC samples
    for (long_int i=1; i<=NM[l]/MULTIMl; i++) {
      
      // draw a sample
      chaos = chaos_draw();
      
      // update chaos on the ACCELERATOR
      // TODO: THIS IS VERY SLOW, SHOULD BE TRANSFERED AS A VECTOR
      //#pragma acc update device(chaos)
      
      // initialize initial data and bc
      init_fvm();
      
      // run the simulation while saving the snapshots
      for (int s=0; s<NSAVES+SAVE0; s++) {
        
        // define when simulation should halt
        if (SAVE0)
          t_halt = 1.0 * MAXT / NSAVES * (s);
        else
          t_halt = 1.0 * MAXT / NSAVES * (s+1);
        
        // run simulation until it halts
        run_fvm();
        
        // record the halt time
        if (i == 1) halt_times[s] = t;
        
        // === compute the statistics for halt time
        
        // compute statistical events for ConservedVars and XtraVars
        if (s == NSAVES+SAVE0-1) {
          compute_events (U);
          compute_events (Xtra);
        }
        
        // ConservedVars
        
        #if COMPUTE_VARIANCE
          
          #if COMPUTE_KURTOSIS
          delta.zero();
          delta_n.zero();
          delta_n2.zero();
          term1.zero();
          
          delta += U;
          delta -= mean_U_l[s];
          
          delta_n += delta;
          delta_n /= i;
          
          delta_n2 += delta_n;
          delta_n2 *= delta_n;
          
          term1 += delta;
          term1 *= delta_n;
          term1 *= i-1;
          
          mean_U_l[s] += delta_n;
          
          M4[s] += term1 * delta_n2 * (i*i - 3*i + 3);
          M4[s] += delta_n2 * M2[s] * 6;
          M4[s] -= delta_n * M3[s] * 4;
          
          M3[s] += term1 * delta_n * (i-2);
          M3[s] -= delta_n * M2[s] * 3;
          
          M2[s] += term1;
          
          #else
          
          if (i>1) {
            temp_U_l.zero();
            temp_U_l += U;
            temp_U_l -= mean_U_l[s]/(i-1);
            temp_U_l.sqr();
            temp_U_l *= (1 - 1.0/i);
            var_U_l[s] += temp_U_l;
          }
          
          #endif
          
        #endif
        
        #if !COMPUTE_KURTOSIS
        mean_U_l[s] += U;
        #endif
        
        // XtraVars
        
        #if COMPUTE_VARIANCE
        if (i>1) {
          temp_X_l.zero();
          temp_X_l += Xtra;
          temp_X_l -= mean_X_l[s]/(i-1);
          temp_X_l.sqr();
          temp_X_l *= (1 - 1.0/i);
          var_X_l[s] += temp_X_l;
        }
        #endif
        mean_X_l[s] += Xtra;
        
        #if DOWNSAMPLE
        if (l != L) {
          
          // all needed samples are in master sampler
          if (ACTIVE_MULTIMl <=1 && multi_samplers_root && i == NM[l+1]) {
            mean_U_l_downsampled [s] += mean_U_l [s];
            mean_X_l_downsampled [s] += mean_X_l [s];
            #if COMPUTE_VARIANCE
            var_U_l_downsampled  [s] += var_U_l  [s];
            var_X_l_downsampled  [s] += var_X_l  [s];
            #endif
          }
          
          // all needed samples are in first ACTIVE_MULTIMl samplers
          else if (multi_rank_samplers < ACTIVE_MULTIMl && i == NM[l]/MULTIMl) {
            mean_U_l_downsampled [s] += mean_U_l [s];
            mean_X_l_downsampled [s] += mean_X_l [s];
            #if COMPUTE_VARIANCE
            var_U_l_downsampled  [s] += var_U_l  [s];
            var_X_l_downsampled  [s] += var_X_l  [s];
            #endif
          }
        }
        #endif
      }
    }
    
    #if COMPUTE_KURTOSIS
    for (int s=0; s<NSAVES+SAVE0; s++) {
      var_U_l [s] += M4[s];
      var_U_l [s] /= NM[l] - 1;
    }
    #endif
    
    // all samplers took part
    ACTIVE_MULTIMl = MULTIMl;
  }
  
  // reuse computed samples
  #if DOWNSAMPLE
  else {
    // samplers that worked on 'downsampled' set of samples
    ACTIVE_MULTIMl = NM[l] / (NM[l-1]/MULTI[M][l-1]);
    
    #if COMPUTE_VARIANCE
    // partition size of the previous level
    MULTIM_PARTITION_SIZE = NM[l-1]/MULTIMl;
    #endif
    
    if ( multi_rank_samplers < max (1, ACTIVE_MULTIMl) ) {
      for (int s=0; s<NSAVES+SAVE0; s++) {
        mean_U_l [s] += mean_U_l_downsampled [s];
        mean_X_l [s] += mean_X_l_downsampled [s];
        #if COMPUTE_VARIANCE
        var_U_l  [s] += var_U_l_downsampled  [s];
        var_X_l  [s] += var_X_l_downsampled  [s];
        #endif
      }
    }
  }
  #endif
  
  // Send and combine the results (mean_U_l, mean_X_l, var_U_l, and var_X_l) of all Samplers of this level to the multi_samplers_root
  multi_combine_samples();
  
  if (multi_samplers_root) {
    // loop over all timesteps
    for (int s=0; s<NSAVES+SAVE0; s++) {
      
      #if !COMPUTE_KURTOSIS
      
      // computing mean & variance
      if (NM[l]>0) // A level could be lost completely
      {
        mean_U_l [s] /= NM[l];
        mean_X_l [s] /= NM[l];
      }
      
      #if COMPUTE_VARIANCE
      var_U_l [s] /= NM[l] - 1;
      var_X_l [s] /= NM[l] - 1;
      #endif
      
      #endif
      
      // for stabilization of ML-part with exp-log
      #if NUMERICAL_STABILIZATION
      
      mean_U_l [s] .exp();
      mean_X_l [s] .exp();
      #if COMPUTE_VARIANCE
      var_U_l  [s] .exp();
      var_X_l  [s] .exp();
      #endif
    
      // multiplying/dividing the final mean & variance to/by the details
      switch (action) {
        case '+' :
          mean_U [s] *= mean_U_l[s];
          mean_X [s] *= mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] *= var_U_l[s];
          var_X [s] *= var_X_l[s];
          #endif
          break;
        
        case '-' :
          mean_U [s] /= mean_U_l[s];
          mean_X [s] /= mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] /= var_U_l[s];
          var_X [s] /= var_X_l[s];
          #endif
      }
      
      // NUMERICAL_STABILIZATION=0
      #else
      // adding/subtracting details to/from the final mean & variance
      switch (action) {
        case '+' :
          mean_U [s] += mean_U_l[s];
          mean_X [s] += mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] += var_U_l[s];
          var_X [s] += var_X_l[s];
          #endif
          break;
        
        case '-' :
          mean_U [s] -= mean_U_l[s];
          mean_X [s] -= mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] -= var_U_l[s];
          var_X [s] -= var_X_l[s];
          #endif
      }
      
      // NUMERICAL_STABILIZATION
      #endif
    }
  }
}

void run_first_difference_component () {
  
  int l = CURRENT_LEVEL;
  
  // no special correlation is needed
  UPSCALE = 0;
  
  display_level_info(l,'+');
  
  LEVEL_INIT_HOOK
  
  if (DOWNSAMPLE) {
    // generate RNG and chaos, use seed from level 0
    rng_and_chaos_init (l, 0);
  }
  else {
    // generate RNG and chaos, use seed from level l
    rng_and_chaos_init (l, l);
  }
    
  //#pragma acc data create(mean_U_l[0:NSAVES+SAVE0],mean_U_l[0:NSAVES+SAVE0]) create(U,Xtra,Source,Flux[0:NDIR])
  
  // get MC estimate of the first term for the current level
  run_samples (l, '+', true);
  
  LEVEL_EXIT_HOOK
  
  display_level_done(l,'+');
}

void run_second_difference_component_type_1 () {
  
  int l = CURRENT_LEVEL;
  
  // second difference component must be coupled (have exaclty same realizations) with the corresponding finer level
  UPSCALE = 1;
  
  // if level is not the finest one, get MC estimate for the E_{M_{l+1}}[u_l]
  if ( l != L ) {
    
    display_level_info(l+1,'-');
    
    LEVEL_INIT_HOOK
    
    if ( DOWNSAMPLE ) {
      // reuse existing results (no additional computing!) to get MC estimate of the second term for the upper level
      run_samples (l+1, '-', false);
    }
    else {
      // generate RNG and chaos for the upper level (use seed from level l+1)
      rng_and_chaos_init (l+1, l+1);
  
      // get MC estimate of the second term for the upper level
      run_samples (l+1, '-', true);
    }
    
    LEVEL_EXIT_HOOK
    
    display_level_done(l+1,'-');
  }
}

void run_second_difference_component_type_2 () {
    
    int l = CURRENT_LEVEL;
    
    // second difference component must be coupled (have exaclty same realizations) with the corresponding finer level
    UPSCALE = 1;
    
    // if level is not the coarsest one, get MC estimate for the E_{M_l}[u_{l-1}]
    if ( l != 0 ) {
      // initialize to the mesh level l-1
      level_init(l-1);
      multi_level_setup();
      
      display_level_info(l,'-');
      
      LEVEL_INIT_HOOK
      
      if (DOWNSAMPLE) {
        // generate RNG and chaos, use seed from level 0
        rng_and_chaos_init (l, 0);
      }
      else {
        // generate RNG and chaos, use seed from level l
        rng_and_chaos_init (l, l);
      }
      
      // get MC estimate of the second term
      run_samples (l, '-', true);
      
      LEVEL_EXIT_HOOK
      
      // restore level setup
      // required even for l=LEVEL_END, to ensure compatibility with multi_combine_levels()
      level_init(l);
      
      display_level_done(l,'-');
    }
}

void run_mlmc() {
  for (int l=LEVEL_BEGIN; l<=LEVEL_END; l++) {
    
    // initialize to the mesh level l
    level_init(l);
    multi_level_init();
    multi_level_setup();
    
    run_first_difference_component ();
    
    #if LOAD_BALANCING_TYPE == 1 || MULTI_TOTAL == 1
    run_second_difference_component_type_1 ();
    #endif
    
    #if LOAD_BALANCING_TYPE == 2
    run_second_difference_component_type_2 ();
    #endif
  }
}

void multi_combine() {
  #if L != 0
  #if WRITE_DATAFILES
  multi_combine_levels();
  #endif
  #endif
  #if HISTOGRAMS
  multi_combine_hist_data();
  #endif
}

void write_datafiles() {
  #if WRITE_DATAFILES
  // saving data
  if ( multi_samplers_root && LEVEL_END == L ) {
    
    #ifndef QUIET
    if (multi_main) std::cout << " :: Writing datafiles..." << std::endl;
    #endif
    
    // saving mean
    Saver saver_mean("mean");
    saver_mean.snapshots();
    
    // saving variance (if computed)
    #if COMPUTE_VARIANCE
    Saver saver_var("variance");
    saver_var.snapshots();
    #endif
    
    #ifndef QUIET
    if (multi_main) std::cout << "  : -> done." << std::endl;
    #endif
  }
  #endif
  
  #if HISTOGRAMS
  // saving HIST_DATA
  #ifndef QUIET
  if (multi_main) std::cout << " :: Writing histogram data..." << std::endl;
  #endif
  save_hist_data();
  #ifndef QUIET
  if (multi_main) std::cout << "  : -> done." << std::endl;
  #endif
  #endif
}
