
#ifndef MLMC_H
#define MLMC_H

#include <math.h>
#include <vector>

#include "config.h"
#include "equation.h"

// === MLMC variables and volumes

// effective doubled order: for solutions with shocks, effective order is halved and bounded by 1
const int EFFECTIVE_DOUBLED_ORDER = SMOOTH_SOLUTION ? 2*ORDER : min(2,ORDER);

// factor for number of samples in each succesive level, i.e. M[l] = FACTOR * M[l+1]
const int FACTOR = 1 << EFFECTIVE_DOUBLED_ORDER;

// factor for work in each succesive level, i.e. WORK[l+1] = WORK_FACTOR * WORK[l]
const int WORK_FACTOR = 1 << ( ACTIVE_DIRS+1 - EFFECTIVE_DOUBLED_ORDER >= 0 ? ACTIVE_DIRS+1 - EFFECTIVE_DOUBLED_ORDER : 0 );

// total number of samples (from all levels)
const long_int NMT = ( ML * ( 1 - ( 1 << (EFFECTIVE_DOUBLED_ORDER * (L+1)) ) ) / ( 1 - FACTOR ) ) * 2  - ML * (1 << (EFFECTIVE_DOUBLED_ORDER*L));

// resolution level at the coarsest mesh
const int L0X = HAVE_DIR(X) ? (int) floor ( log2(NX) ) - L : 0;
const int L0Y = HAVE_DIR(Y) ? (int) floor ( log2(NY) ) - L : 0;
const int L0Z = HAVE_DIR(Z) ? (int) floor ( log2(NZ) ) - L : 0;

// === global variables

extern int N[NDIR][L+1];                // number of cells for each level
extern long_int NM[L+1];                // number of samples for each level

// main containers for mean and variance
extern volume <ConservedVars, 1> mean_U [NSAVES+SAVE0];
extern volume <XtraVars, 1>      mean_X [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
extern volume <ConservedVars, 1> var_U  [NSAVES+SAVE0];
extern volume <XtraVars, 1>      var_X  [NSAVES+SAVE0];
#endif

// level-dependent containers for mean and variance
extern volume <ConservedVars, 1> mean_U_l [NSAVES+SAVE0];
extern volume <XtraVars, 1>      mean_X_l [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
extern volume <ConservedVars, 1> var_U_l  [NSAVES+SAVE0];
extern volume <XtraVars, 1>      var_X_l  [NSAVES+SAVE0];
#endif

// temporary volumes for computational efficiency
#if COMPUTE_VARIANCE
extern volume <ConservedVars, 1> temp_U_l;
extern volume <XtraVars,      1> temp_X_l;
#endif

// extra quantities of interest (e.g. max wave speed) from each sample
#if HISTOGRAMS
extern std::vector<real> HIST_DATA [L+1];
#endif

// === MLMC routines

// computes MLMC paramaters such as N[*][*], NM[*], MULTI[:][:], MULTI_TOTAL, LOWEST_MULTI_LEVEL, etc.
// should be executed before any other routine is used
void mlmc_init ();

// run compatibility checks
// should be executed after mlmc_init()
void run_compatibility_checks ();

// display information such as number of cores, levels, coarsest level, etc.
// should be executed after mlmc_init()
void display_info ();

// initializes main volumes for mean and variance
// should be executed after mlmc_init()
void volumes_init ();

// setup meshes for level l
// should be executed after volumes_init()
void level_init (int l);

// display information such as resolution, number of samples, number of cores, etc.
// should be executed after level_init()
// 'l' indicates level from which the number of samples is taken
// 'action' indicates which difference component is active, '+' or '-'
void display_level_info(int l, char action);

// display that the level has finished
// input parameters as for display_level_info()
void display_level_done(int l, char action);

// get the pre-set seed for level l
unsigned int getSeed (unsigned int level);

// generate samples in _each_ node with an appropriate offset
// 'l_samples' indicates from which level number of samples should be taken
// 'l_seed' indicates from which level the seed should be taken
void rng_and_chaos_init (int l_samples, int l_seed );

// run NM[l] samples; level 'l' indicated from which level the number of samples should be taken; it does NOT incidate mesh resolution
// 'action' indicates whether computed ensemble (of this level) should be added or subtracted from the global ensemble
// 'compute' indicates whether samples should be computed or stored samples should be reused (when DOWNSAMPLING=1)
void run_samples (int l, char action, bool compute);

// run MC samples for the first component in the coupled differences
void run_first_difference_component ();

// run MC samples for the second component in the coupled differences (LOAD_BALANCING_TYPE=1)
// NM[l] samples from the COARSER level pair are computed on the SAME resolution (set by level_init())
// shoould be also used in serial mode (with no paralellization)
void run_second_difference_component_type_1 ();

// run MC samples for the second component in the coupled differences (LOAD_BALANCING_TYPE=2)
// NM[l] samples from the SAME level pair are computed on the COARSER resolution (set by level_init())
// shoud NOT be used in serial mode (with no paralellization)
void run_second_difference_component_type_2 ();

// run MLMC algorithm: compute MC estimates for levels l, where: LEVEL_BEGIN <= l <= LEVEL_END
void run_mlmc();

// combine data from all cores
void multi_combine();

// write datafiles
void write_datafiles();

#endif
