
#include "multi-mpi-helper.h"
#include "multi.h"
#include "mlmc.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include "mpi.h"

// timer for wall-clock time
double wallclock_timer;
double wallclock_runtime = 0;

// timer for wall-clock time of DDM-specific MPI routines
double wallclock_timer_mpi_ddm;
double wallclock_runtime_mpi_ddm = 0;

// timer for wall-clock time of MC-specific MPI routines
double wallclock_timer_mpi_mc;
double wallclock_runtime_mpi_mc = 0;

// timer for wall-clock time of MLMC-specific MPI routines
double wallclock_timer_mpi_mlmc;
double wallclock_runtime_mpi_mlmc = 0;

// timer for wall-clock time of idling due to load balancing
double wallclock_timer_mpi_idle;
double wallclock_runtime_mpi_idle = 0;

void multi_run_compatibility_checks () {
  
  // check for condition s <= (d+1)/2
  if (L != 0 && EFFECTIVE_DOUBLED_ORDER > ACTIVE_DIRS + 1 && multi_main) {
    std::cout << " :: ERROR: 2s <= d+1 is violated: " << EFFECTIVE_DOUBLED_ORDER << " > " << ACTIVE_DIRS + 1 << std::endl;
    std::cout << "  : Solution is expected to be ";
    if (!SMOOTH_SOLUTION) std::cout << "non-";
    std::cout << "smooth. (SMOOTH_SOLUTION=" << SMOOTH_SOLUTION << ")" << std::endl;
    multi_finalise();
    std::exit(1);
  }
}

void start_timer () {
  MPI_Barrier (MPI_COMM_WORLD);
  if (multi_main) wallclock_timer = MPI_Wtime();
}

void stop_timer () {
  
  // computing global routine-specific MPI timers
  
  double wallclock_runtime_global_mpi_ddm;
  double wallclock_runtime_global_mpi_mc;
  double wallclock_runtime_global_mpi_mlmc;
  double wallclock_runtime_global_mpi_idle;
  
  MPI_Reduce (&wallclock_runtime_mpi_ddm,  &wallclock_runtime_global_mpi_ddm,  1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce (&wallclock_runtime_mpi_mc,   &wallclock_runtime_global_mpi_mc,   1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce (&wallclock_runtime_mpi_mlmc, &wallclock_runtime_global_mpi_mlmc, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce (&wallclock_runtime_mpi_idle, &wallclock_runtime_global_mpi_idle, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  
  if (multi_main) {
    
    std::cout << std::endl;
    
    wallclock_runtime = MPI_Wtime() - wallclock_timer;
    
    std::cout << " :: Wall-clock runtime          : ";
    print_runtime (wallclock_runtime);
    std::cout << std::endl;
    
    double wallclock_runtime_global = wallclock_runtime * MULTI_TOTAL;
    
    std::cout << " :: Wall-clock runtime (global) : ";
    print_runtime (wallclock_runtime_global);
    std::cout << std::endl;
    
    double wallclock_runtime_global_mpi = 0;
    wallclock_runtime_global_mpi += wallclock_runtime_global_mpi_ddm;
    wallclock_runtime_global_mpi += wallclock_runtime_global_mpi_mc;
    wallclock_runtime_global_mpi += wallclock_runtime_global_mpi_mlmc;
    wallclock_runtime_global_mpi += wallclock_runtime_global_mpi_idle;
    
    std::cout << " :: MPI overhead                : ";
    print_runtime (wallclock_runtime_global_mpi);
    std::cout << std::endl;
    
    std::cout << "  :   DDM:  " << round(1000 * wallclock_runtime_global_mpi_ddm  / wallclock_runtime_global)/10 << "\%";
    std::cout << std::endl;
    
    std::cout << "  :   MC:   " << round(1000 * wallclock_runtime_global_mpi_mc   / wallclock_runtime_global)/10 << "\%";
    std::cout << std::endl;
    
    std::cout << "  :   MLMC: " << round(1000 * wallclock_runtime_global_mpi_mlmc / wallclock_runtime_global)/10 << "\%";
    std::cout << std::endl;
    
    std::cout << "  :   Idle: " << round(1000 * wallclock_runtime_global_mpi_idle / wallclock_runtime_global)/10 << "\%";
    std::cout << std::endl;
    
    double mpi_efficiency = 1 - wallclock_runtime_global_mpi/wallclock_runtime_global;
    if (mpi_efficiency < 0) mpi_efficiency = 0;
    
    std::cout << " :: MPI efficiency              : " << round(1000*mpi_efficiency)/10 << "\%" << std::endl;
    
    std::ofstream runtimef (RUNTIME_FILE);
    runtimef << "wallclock_runtime="                 << wallclock_runtime                 << std::endl;
    runtimef << "wallclock_runtime_global="          << wallclock_runtime_global          << std::endl;
    runtimef << "wallclock_runtime_global_mpi_ddm="  << wallclock_runtime_global_mpi_ddm  << std::endl;
    runtimef << "wallclock_runtime_global_mpi_mc="   << wallclock_runtime_global_mpi_mc   << std::endl;
    runtimef << "wallclock_runtime_global_mpi_mlmc=" << wallclock_runtime_global_mpi_mlmc << std::endl;
    runtimef << "wallclock_runtime_global_mpi_idle=" << wallclock_runtime_global_mpi_idle << std::endl;
    runtimef << "wallclock_runtime_global_mpi="      << wallclock_runtime_global_mpi      << std::endl;
    runtimef << "mpi_efficiency="                    << mpi_efficiency                    << std::endl;
    runtimef << "cores="                             << MULTI_TOTAL                       << std::endl;
    runtimef.flush();
  }
}

void resume_timer_mpi_ddm () {
  wallclock_timer_mpi_ddm = MPI_Wtime();
}

void halt_timer_mpi_ddm () {
  wallclock_runtime_mpi_ddm += MPI_Wtime() - wallclock_timer_mpi_ddm;
}

void resume_timer_mpi_mc () {
  wallclock_timer_mpi_mc = MPI_Wtime();
}

void halt_timer_mpi_mc () {
  wallclock_runtime_mpi_mc += MPI_Wtime() - wallclock_timer_mpi_mc;
}

void resume_timer_mpi_mlmc () {
  wallclock_timer_mpi_mlmc = MPI_Wtime();
}

void halt_timer_mpi_mlmc () {
  wallclock_runtime_mpi_mlmc += MPI_Wtime() - wallclock_timer_mpi_mlmc;
}

void resume_timer_mpi_idle () {
  wallclock_timer_mpi_idle = MPI_Wtime();
}

void halt_timer_mpi_idle () {
  wallclock_runtime_mpi_idle += MPI_Wtime() - wallclock_timer_mpi_idle;
}

void multi_set_first_and_last_level () {
  int level = L;
  int nodes = MULTI[M][level] * MULTI[X][level] * MULTI[Y][level] * MULTI[Z][level];
  
  while (multi_rank >= nodes) {
    level--;
    nodes += MULTI[M][level] * MULTI[X][level] * MULTI[Y][level] * MULTI[Z][level];
  }
  
  // if there is no parallelisation after this level, i.e. this level includes all remaining lower levels
  if (level == LOWEST_MULTI_LEVEL) {
    LEVEL_BEGIN = 0;
    LEVEL_END   = level;
  }
  
  // else this level is paralellised as usual
  else {
    LEVEL_BEGIN = level;
    LEVEL_END   = level;
  }
}

int get_nodes_down_to_level (int l) {
  int nodes = 0;
  int level = L;
  
  while (level > l) {
    nodes += MULTI[M][level] * MULTI[X][level] * MULTI[Y][level] * MULTI[Z][level];
    level--;
  }
  
  return nodes;
}

void get_group_nodes (const int multix, const int multiy, const int multiz, int *group_nodes) {
  
  int needed_multix, needed_multiy, needed_multiz;
  
  for (int l=L; l>=0; l--) {
    needed_multix = multix / (MULTIX / MULTI[X][l]);
    needed_multiy = multiy / (MULTIY / MULTI[Y][l]);
    needed_multiz = multiz / (MULTIZ / MULTI[Z][l]);
    
    group_nodes[L-l] =  get_nodes_down_to_level(l);
    group_nodes[L-l] += needed_multiz*MULTI[X][l]*MULTI[Y][l];
    group_nodes[L-l] += needed_multiy*MULTI[X][l];
    group_nodes[L-l] += needed_multix;
  }
}
