
#include "multi.h"

#include "mlmc.h"
#include "multi_mpi.h"
#include "multi-mpi-helper.h"
#include "chaos.h"

#include "mpi.h"
#include <cstdlib>

///::title:: MPI parallelisation
///// Number of zones in each direction (each zone gets its own core)
///::vars:: MULTIX=1 MULTIY=1 MULTIZ=1
///// Number of separate groups of cores for sampling
///::vars:: MULTIM=1
///::consts:: BC_REPEAT_X=0 BC_REPEAT_Y=0 BC_REPEAT_Z=0
///::depends:: multi-mpi-helper

// FVM - MPI

int mpir, mpinbc, mpi_BUFSIZE;
int mpibc[NDIR*2], mpibcr[NDIR*2], mpi_bc_rbuf[NDIR*2];
ConservedVars *mpi_rbuf[NDIR*2], *mpi_sbuf[NDIR*2];

int rank_domain;

MPI_Request req_hold;

// MLMC - MPI

MPI_Group group_main;

MPI_Group group_levels;
MPI_Comm   comm_levels;
MPI_Group group_level;
MPI_Comm   comm_level;
MPI_Group group_samplers;
MPI_Comm   comm_samplers;
MPI_Group group_domain;
MPI_Comm   comm_domain;

int rank_level;

Chaos *mpi_chaos_hold_buf;

int group_nodes [MULTIX][MULTIY][MULTIZ][L+1];

// routines

void multi_init(int argc, char **argv)
{
  int n;
  MPI_Init(&argc, &argv);
  
  MPI_Comm_rank(MPI_COMM_WORLD, &multi_rank);
  multi_main = (multi_rank == 0);
  
  MPI_Comm_size(MPI_COMM_WORLD, &n);
  if (n != MULTI_TOTAL) {
    if (multi_main) std::cerr << " ERROR: inconsistent number of processes" << std::endl;
    MPI_Finalize();
    std::exit(1);
  }
  
  if (multi_main) std::cout << " :: INFO: number of cores: " << MULTI_TOTAL << std::endl;
  
  MPI_Comm_group (MPI_COMM_WORLD, &group_main);
  
  multi_set_first_and_last_level ();
  
  // create comm_level and group_level
  
  int level = LEVEL_END;
  
  int first_rank = get_nodes_down_to_level (level);
  int last_rank = get_nodes_down_to_level (level - 1) - 1;
  int stride = 1;
  int range [1][3] = {{first_rank, last_rank, stride}};
  
  MPI_Group_range_incl ( group_main, 1, range, &group_level );
  MPI_Comm_create ( MPI_COMM_WORLD, group_level, &comm_level );
  
  MPI_Comm_rank(comm_level, &rank_level);
  multi_level_root = (rank_level == 0);
  
  // precomputing group_nodes
  
  for (int mx=0; mx<MULTIX; mx++)
    for (int my=0; my<MULTIY; my++)
      for (int mz=0; mz<MULTIZ; mz++)
        get_group_nodes(mx, my, mz, group_nodes[mx][my][mz]);
  
  // create comm_levels and group_levels
  
  int size_levels = L+1 - LOWEST_MULTI_LEVEL;
  int nodes_levels [size_levels];
    
  for (int i=0; i<size_levels; i++)
    nodes_levels[i] = get_nodes_down_to_level (L-i);
  
  MPI_Group_incl(group_main, size_levels, &nodes_levels[0], &group_levels);
  MPI_Comm_create ( MPI_COMM_WORLD, group_levels, &comm_levels );
}

void multi_level_init () {
  
  resume_timer_mpi_ddm();
  
  int first_rank, last_rank, stride;
  
  // create domain communicators
  
  first_rank = rank_level - (rank_level % MULTINl);
  last_rank = first_rank + MULTINl - 1;
  stride = 1;
  int range_domain [1][3] = {{first_rank, last_rank, stride}};
  MPI_Group_range_incl ( group_level, 1, range_domain, &group_domain );
  MPI_Comm_create ( comm_level, group_domain, &comm_domain );
  
  MPI_Comm_rank(comm_domain, &rank_domain);
  multi_domain_root = (rank_domain == 0);
  
  // create samplers communicators
  
  first_rank = rank_domain;
  last_rank = first_rank + (MULTIMl - 1)*MULTINl;
  stride = MULTINl;
  int range_samplers [1][3] = {{first_rank, last_rank, stride}};
  MPI_Group_range_incl ( group_level, 1, range_samplers, &group_samplers );
  MPI_Comm_create ( comm_level, group_samplers, &comm_samplers ); 
  
  int rank_samplers;
  MPI_Comm_rank(comm_samplers, &rank_samplers);
  multi_samplers_root = (rank_samplers == 0);
  multi_rank_samplers = rank_samplers;
  
  // remaining
  
  multiz = rank_domain / (MULTIXl*MULTIYl);
  multiy = (rank_domain / MULTIXl) % MULTIYl;
  multix = rank_domain % MULTIXl;
  
  mpir = rank_domain;
  if (BC_REPEAT_X && MULTIXl > 1) {
    mpibc[X] = 1;
    mpibcr[X] = rank_domain - rank_domain % MULTIXl + (rank_domain + MULTIXl - 1) % MULTIXl ;
    mpibc[X+NDIR] = 1;
    mpibcr[X+NDIR] = rank_domain - rank_domain % MULTIXl + (rank_domain + 1) % MULTIXl;
  } else {
    mpibc[X] = (multix > 0);
    if (mpibc[X]) mpibcr[X] = rank_domain - 1;
    mpibc[X+NDIR] = (multix < MULTIXl-1);
    if (mpibc[X+NDIR]) mpibcr[X+NDIR] = rank_domain + 1;
  }
  const int MXY = MULTIXl*MULTIYl;
  if (BC_REPEAT_Y && MULTIYl > 1) {
    mpibc[Y] = 1;
    mpibcr[Y] = rank_domain - rank_domain % MXY + (rank_domain + MXY - MULTIXl) % MXY;
    mpibc[Y+NDIR] = 1;
    mpibcr[Y+NDIR] = rank_domain - rank_domain % MXY + (rank_domain + MULTIXl) % MXY;
  } else {
    mpibc[Y] = (multiy > 0);
    if (mpibc[Y]) mpibcr[Y] = rank_domain - MULTIXl;
    mpibc[Y+NDIR] = (multiy < MULTIYl-1);
    if (mpibc[Y+NDIR]) mpibcr[Y+NDIR] = rank_domain + MULTIXl;
  }
  const int MXYZ = MXY*MULTIZl;
  if (BC_REPEAT_Z && MULTIZl > 1) {
    mpibc[Z] = 1;
    mpibcr[Z] = (rank_domain + MXYZ - MXY) % MXYZ;
    mpibc[Z+NDIR] = 1;
    mpibcr[Z+NDIR] = (rank_domain + MXY) % MXYZ;
  } else {
    mpibc[Z] = (multiz > 0);
    if (mpibc[Z]) mpibcr[Z] = rank_domain - MXY;
    mpibc[Z+NDIR] = (multiz < MULTIZl-1);
    if (mpibc[Z+NDIR]) mpibcr[Z+NDIR] = rank_domain + MXY;
  }
  mpinbc = mpibc[X]+mpibc[Y]+mpibc[Z]+mpibc[X+NDIR]+mpibc[Y+NDIR]+mpibc[Z+NDIR];
  
  halt_timer_mpi_ddm();
}

void multi_level_setup () {
  
  resume_timer_mpi_ddm();
  
  // garbage collection from from previous setups
  for (int d = 0; d < 2*NDIR; d++) {
    delete [] mpi_rbuf[d];
    delete [] mpi_sbuf[d];
  }
  
  if (NZl == 1) {
    if (NYl == 1) mpi_BUFSIZE = 1;
    else if (NYl > NXl) mpi_BUFSIZE = NYl;
    else mpi_BUFSIZE = NXl;
  } else if (NZl > NYl) {
    if (NXl > NYl) mpi_BUFSIZE = NXl*NZl;
    else mpi_BUFSIZE = NYl*NZl;
  } else if (NXl > NZl) mpi_BUFSIZE = NXl*NYl;
  else mpi_BUFSIZE = NYl*NZl;
  mpi_BUFSIZE *= GHOST_SIZE;
  for (int d = 0; d < 2*NDIR; d++) {
    mpi_rbuf[d] = new ConservedVars[mpi_BUFSIZE];
    mpi_sbuf[d] = new ConservedVars[mpi_BUFSIZE];
  }
  
  halt_timer_mpi_ddm();
}

real multi_dt(real dt)
{
  resume_timer_mpi_ddm();
  real dtr;
  MPI_Allreduce(&dt, &dtr, 1, mpi_REAL, MPI_MIN, comm_domain);
  
  halt_timer_mpi_ddm();
  
  return dtr;
}

void multi_bc()
{
  resume_timer_mpi_ddm();
  
  MPI_Status s;
  MPI_Request req[2*NDIR];
  // copy all internal borders into the send buffers
  if (NXl > 1 && MULTIXl > 1) {
    for (int i = 0; i < GX; i++) {
      for (int j = 0; j < NYl; j++) {
        for (int k = 0; k < NZl; k++) {
          if (mpibc[X]) mpi_sbuf[X][((i*NYl) + j)*NZl + k] = U[i][j][k];
          if (mpibc[X+NDIR])
            mpi_sbuf[X+NDIR][((i*NYl) + j)*NZl + k] = U[NXl-1-i][j][k];
        }
      }
    }
  }
  if (NYl > 1 && MULTIYl > 1) {
    for (int i = 0; i < NXl; i++) {
      for (int j = 0; j < GY; j++) {
        for (int k = 0; k < NZl; k++) {
          if (mpibc[Y]) mpi_sbuf[Y][((i*GY) + j)*NZl + k] = U[i][j][k];
          if (mpibc[Y+NDIR])
            mpi_sbuf[Y+NDIR][((i*GY) + j)*NZl + k] = U[i][NYl-1-j][k];
        }
      }
    }
  }
  if (NZl > 1 && MULTIZl > 1) {
    for (int i = 0; i < NXl; i++) {
      for (int j = 0; j < NYl; j++) {
        for (int k = 0; k < GZ; k++) {
          if (mpibc[Z]) mpi_sbuf[Z][((i*NYl) + j)*GZ + k] = U[i][j][k];
          if (mpibc[Z+NDIR])
            mpi_sbuf[Z+NDIR][((i*NYl) + j)*GZ + k] = U[i][j][NZl-1-k];
        }
      }
    }
  }
  // start the sending of the boundary data
  for (int i = 0; i < 2*NDIR; i++) {
    if (mpibc[i]) {
      int size = SIZE_CV*NXl*NYl*NZl*GHOST_SIZE / SIZE(i%3);
      MPI_Isend(mpi_sbuf[i], size, mpi_REAL, mpibcr[i], (i+3)%6,
                comm_domain, &req[i]);
    }
  }
  // receive all boundary data  
  for (int i = 0; i < mpinbc; i++) {
    MPI_Recv(mpi_rbuf[i], SIZE_CV*mpi_BUFSIZE, mpi_REAL, MPI_ANY_SOURCE, MPI_ANY_TAG,
            comm_domain, &s);
    mpi_bc_rbuf[s.MPI_TAG] = i;
  }
  // confirm the completion of the sends
  for (int i = 0; i < 2*NDIR; i++)
    if (mpibc[i]) MPI_Wait(&req[i], &s);
  
  halt_timer_mpi_ddm();
}

void multi_combine_samples () {
  
  resume_timer_mpi_mc();
  
  if (ACTIVE_MULTIMl > 1) {
    
    // need to combine all snapshots
    for(int s = 0; s<NSAVES+SAVE0; s++) {
      
      int SIZE_DATA = mean_U_l[s].VNX * mean_U_l[s].VNY * mean_U_l[s].VNZ;
      
      // send buffer
      void *mpi_send_buf;
      if (multi_samplers_root) mpi_send_buf = MPI_IN_PLACE;
      
      int count, tag;
      
      // === ConservedVars
      
      count = SIZE_CV * SIZE_DATA;
      tag   = 0;
      
      // mean
      #if COMPUTE_VARIANCE
      if ( !multi_samplers_root && multi_rank_samplers < ACTIVE_MULTIMl )
        MPI_Send (mean_U_l[s].data, count, mpi_REAL, 0, tag, comm_samplers);
      
      #else
      if ( !multi_samplers_root ) mpi_send_buf = mean_U_l[s].data;      
      MPI_Reduce ( mpi_send_buf, mean_U_l[s].data, count, mpi_REAL, MPI_SUM, 0, comm_samplers );
      
      #endif
      
      // variance
      #if COMPUTE_VARIANCE
      // update mean while updating variance as well
      if (multi_samplers_root) {
        
        volume <ConservedVars, 1> temp2_U_l (NXl, NYl, NZl);
        
        for (int i=1; i<ACTIVE_MULTIMl; i++) {
          MPI_Recv (temp_U_l.data, count, mpi_REAL, i, tag, comm_samplers, MPI_STATUS_IGNORE);
          
          temp2_U_l.zero();
          temp2_U_l += temp_U_l;
          temp2_U_l /= MULTIM_PARTITION_SIZE;
          temp2_U_l -= mean_U_l[s] / (i*MULTIM_PARTITION_SIZE);   // NOT COOL
          temp2_U_l.sqr();
          temp2_U_l *= ( i / (i+1.0) ) * MULTIM_PARTITION_SIZE;
          
          var_U_l[s] += temp2_U_l;
          
          mean_U_l[s] += temp_U_l;
        }
      }
      
      // add main contribution for the variance
      if ( !multi_samplers_root ) mpi_send_buf = var_U_l[s].data;
      MPI_Reduce ( mpi_send_buf, var_U_l[s].data, count, mpi_REAL, MPI_SUM, 0, comm_samplers );
      #endif
      
      // === XtraVars
      
      count = SIZE_XV * SIZE_DATA;
      tag   = 1;
      
      // mean
      #if COMPUTE_VARIANCE
      if ( !multi_samplers_root  && multi_rank_samplers < ACTIVE_MULTIMl )
        MPI_Send (mean_X_l[s].data, count, mpi_REAL, 0, tag, comm_samplers);
      
      #else
      if ( !multi_samplers_root ) mpi_send_buf = mean_X_l[s].data;      
      MPI_Reduce ( mpi_send_buf, mean_X_l[s].data, count, mpi_REAL, MPI_SUM, 0, comm_samplers );
      
      #endif
      
      // variance
      #if COMPUTE_VARIANCE
      // update mean while updating variance as well
      if (multi_samplers_root) {
        
        volume <XtraVars, 1> temp2_X_l (NXl, NYl, NZl);
        
        for (int i=1; i<ACTIVE_MULTIMl; i++) {
          MPI_Recv (temp_X_l.data, count, mpi_REAL, i, tag, comm_samplers, MPI_STATUS_IGNORE);
          
          temp2_X_l.zero();
          temp2_X_l += temp_X_l;
          temp2_X_l /= MULTIM_PARTITION_SIZE;
          temp2_X_l -= mean_X_l[s] / (i*MULTIM_PARTITION_SIZE);   // need to use memory more efficiently?
          temp2_X_l.sqr();
          temp2_X_l *= ( i / (i+1.0) ) * MULTIM_PARTITION_SIZE;
          
          var_X_l[s] += temp2_X_l;
          
          mean_X_l[s] += temp_X_l;
        }
      }
      
      // add main contribution for the variance
      if ( !multi_samplers_root ) mpi_send_buf = var_X_l[s].data;
      MPI_Reduce ( mpi_send_buf, var_X_l[s].data, count, mpi_REAL, MPI_SUM, 0, comm_samplers );
      #endif
    }
  }
  
  halt_timer_mpi_mc();
}

void multi_combine_levels () {
  
  // determining load imbalance
  
  resume_timer_mpi_idle();
  
  MPI_Barrier (MPI_COMM_WORLD);
  
  halt_timer_mpi_idle();
  
  // combining MLMC levels
  
  resume_timer_mpi_mlmc();
  
  #ifndef QUIET
  if (multi_main) std::cout << " :: Combining samples...";
  #endif
  
  const int TAG_MEAN_CV = 0;
  const int TAG_MEAN_XV = 1;
  const int TAG_VAR_CV  = 2;
  const int TAG_VAR_XV  = 3;
  
  // need to combine all snapshots
  for(int s = 0; s<NSAVES+SAVE0; s++) {
    
    // all information is in sampler_roots
    if (multi_samplers_root) {
      
      // master level nodes RECEIVE data from other levels
      if (LEVEL_BEGIN == L) {
        
        for (int l=LOWEST_MULTI_LEVEL; l<L; l++) {
          
          int source = group_nodes[multix][multiy][multiz][L-l];
          
          // === ConservedVars
          volume<ConservedVars, 1> sub_mean_CV (N[X][l] / (MULTIX / MULTI[X][l]), N[Y][l] / (MULTIY / MULTI[Y][l]), N[Z][l] / (MULTIZ / MULTI[Z][l]));
          #if COMPUTE_VARIANCE
          volume<ConservedVars, 1> sub_var_CV  (N[X][l] / (MULTIX / MULTI[X][l]), N[Y][l] / (MULTIY / MULTI[Y][l]), N[Z][l] / (MULTIZ / MULTI[Z][l]));
          #endif
          
          int count_CV = SIZE_CV * sub_mean_CV.VNX * sub_mean_CV.VNY * sub_mean_CV.VNZ;
          
          MPI_Recv (sub_mean_CV.data, count_CV, mpi_REAL, source, TAG_MEAN_CV, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          // for stabilization of multi-level difference computation with exp-log
          #if NUMERICAL_STABILIZATION
          mean_U [s] *= sub_mean_CV;
          #else
          mean_U [s] += sub_mean_CV;
          #endif
          
          #if COMPUTE_VARIANCE
          MPI_Recv (sub_var_CV.data, count_CV, mpi_REAL, source, TAG_VAR_CV, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          #if NUMERICAL_STABILIZATION
          var_U [s] *= sub_var_CV;
          #else
          var_U [s] += sub_var_CV;
          #endif
          #endif
          
          // === XtraVars
          volume<XtraVars, 1> sub_mean_XV (N[X][l] / (MULTIX / MULTI[X][l]), N[Y][l] / (MULTIY / MULTI[Y][l]), N[Z][l] / (MULTIZ / MULTI[Z][l]));
          #if COMPUTE_VARIANCE
          volume<XtraVars, 1> sub_var_XV  (N[X][l] / (MULTIX / MULTI[X][l]), N[Y][l] / (MULTIY / MULTI[Y][l]), N[Z][l] / (MULTIZ / MULTI[Z][l]));
          #endif
          
          int count_XV = SIZE_XV * sub_mean_XV.VNX * sub_mean_XV.VNY * sub_mean_XV.VNZ;
          
          MPI_Recv (sub_mean_XV.data, count_XV, mpi_REAL, source, TAG_MEAN_XV, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          #if NUMERICAL_STABILIZATION
          mean_X [s] *= sub_mean_XV;
          #else
          mean_X [s] += sub_mean_XV;
          #endif
          
          #if COMPUTE_VARIANCE
          MPI_Recv (sub_var_XV.data, count_XV, mpi_REAL, source, TAG_VAR_XV, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          #if NUMERICAL_STABILIZATION
          var_X [s] *= sub_var_XV;
          #else
          var_X [s] += sub_var_XV;
          #endif
          #endif
        }
      }
      
      // nodes from other levels SEND data to master level nodes
      else {
        
        int N_GROUPSX = (MULTIX / MULTIXl);
        int N_GROUPSY = (MULTIY / MULTIYl);
        int N_GROUPSZ = (MULTIZ / MULTIZl);
        
        volume<ConservedVars, 1> sub_mean_CV [N_GROUPSX][N_GROUPSY][N_GROUPSZ];
        volume<XtraVars, 1>      sub_mean_XV [N_GROUPSX][N_GROUPSY][N_GROUPSZ];
        
        #if COMPUTE_VARIANCE
        volume<ConservedVars, 1> sub_var_CV  [N_GROUPSX][N_GROUPSY][N_GROUPSZ];
        volume<XtraVars, 1>      sub_var_XV  [N_GROUPSX][N_GROUPSY][N_GROUPSZ];
        #endif
        
        int sub_NX = NXl/N_GROUPSX;
        int sub_NY = NYl/N_GROUPSY;
        int sub_NZ = NZl/N_GROUPSZ;
        
        for (int mx=0; mx < N_GROUPSX; mx++) {
          for (int my=0; my < N_GROUPSY; my++) {
            for (int mz=0; mz < N_GROUPSZ; mz++) {
              
              int needed_multix = mx + multix*N_GROUPSX;
              int needed_multiy = my + multiy*N_GROUPSY;
              int needed_multiz = mz + multiz*N_GROUPSZ;
              
              sub_mean_CV[mx][my][mz].reinit (sub_NX, sub_NY, sub_NZ);
              sub_mean_XV[mx][my][mz].reinit (sub_NX, sub_NY, sub_NZ);
              
              #if COMPUTE_VARIANCE
              sub_var_CV [mx][my][mz].reinit (sub_NX, sub_NY, sub_NZ);
              sub_var_XV [mx][my][mz].reinit (sub_NX, sub_NY, sub_NZ);
              #endif
              
              for (int i = 0; i < sub_NX; i++) {
                for (int j = 0; j < sub_NY; j++) {
                  for (int k = 0; k < sub_NZ; k++) {
                    sub_mean_CV[mx][my][mz] [i][j][k] = mean_U [s] [i + mx*sub_NX] [j + my*sub_NY] [k + mz*sub_NZ];
                    sub_mean_XV[mx][my][mz] [i][j][k] = mean_X [s] [i + mx*sub_NX] [j + my*sub_NY] [k + mz*sub_NZ];
                    
                    #if COMPUTE_VARIANCE
                    sub_var_CV [mx][my][mz] [i][j][k] = var_U [s] [i + mx*sub_NX] [j + my*sub_NY] [k + mz*sub_NZ];
                    sub_var_XV [mx][my][mz] [i][j][k] = var_X [s] [i + mx*sub_NX] [j + my*sub_NY] [k + mz*sub_NZ];
                    #endif
                  }
                }
              }
              
              int count_CV = SIZE_CV * sub_mean_CV[mx][my][mz].VNX * sub_mean_CV[mx][my][mz].VNY * sub_mean_CV[mx][my][mz].VNZ;
              
              MPI_Send (sub_mean_CV[mx][my][mz].data, count_CV, mpi_REAL, group_nodes[needed_multix][needed_multiy][needed_multiz][0], TAG_MEAN_CV, MPI_COMM_WORLD);
              #if COMPUTE_VARIANCE
              MPI_Send (sub_var_CV[mx][my][mz].data,  count_CV, mpi_REAL, group_nodes[needed_multix][needed_multiy][needed_multiz][0], TAG_VAR_CV, MPI_COMM_WORLD);
              #endif
              
              int count_XV = SIZE_XV * sub_mean_XV[mx][my][mz].VNX * sub_mean_XV[mx][my][mz].VNY * sub_mean_XV[mx][my][mz].VNZ;
              
              MPI_Send (sub_mean_XV[mx][my][mz].data, count_XV, mpi_REAL, group_nodes[needed_multix][needed_multiy][needed_multiz][0], TAG_MEAN_XV, MPI_COMM_WORLD);
              #if COMPUTE_VARIANCE
              MPI_Send (sub_var_XV[mx][my][mz].data,  count_XV, mpi_REAL, group_nodes[needed_multix][needed_multiy][needed_multiz][0], TAG_VAR_XV, MPI_COMM_WORLD);
              #endif
            }
          }
        }
      }
    }
  }
  
  #ifndef QUIET
  if (multi_main) std::cout << " done." << std::endl;
  #endif
  
  halt_timer_mpi_mlmc();
}

#if HISTOGRAMS
void multi_combine_hist_data () {
    
    resume_timer_mpi_mlmc();
  
    #ifndef QUIET
    if (multi_main) std::cout << " :: Combining histogram data...";
    #endif
    
    // pointer to deal with send buffers
    void *mpi_send_buf;
    
    // post-process data from each subdomain to domain roots
    for (int l=0; l<=L; l++) {
      // setup send buffer
      mpi_send_buf = multi_domain_root ? MPI_IN_PLACE : HIST_DATA[l].data();
      // commence!
      MPI_Reduce ( mpi_send_buf, HIST_DATA[l].data(), HIST_DATA[l].size(), mpi_REAL, MPI_MAX, 0, comm_domain );
    }
    
    // gather data from domain roots to samplers roots
    for (int l=0; l<=L; l++) {
      // reserve space on the samplers roots
      if (multi_domain_root) {
        // determine required sizes
        int hist_data_chunk = HIST_DATA[l].size();
        int hist_data_total = MULTI[M][l] * hist_data_chunk;
        // reserve space on the samplers roots
        if (multi_samplers_root) HIST_DATA[l] .resize( hist_data_total );
        // setup send buffer
        mpi_send_buf = multi_samplers_root ? MPI_IN_PLACE : HIST_DATA[l].data();
        // commence!
        MPI_Gather ( mpi_send_buf, hist_data_chunk, mpi_REAL, HIST_DATA[l].data(), hist_data_chunk, mpi_REAL, 0, comm_samplers );
      }
    }
    
    // gather data from samplers roots to root
    for (int l=0; l<=L; l++) {
      if (multi_domain_root && multi_samplers_root) {
        // determine required sizes
        int hist_data_chunk = HIST_DATA[l].size();
        int hist_data_total = 0;
        MPI_Reduce ( &hist_data_chunk, &hist_data_total, 1, MPI_INT, MPI_SUM, 0, comm_levels );
        // sizes of data to be received from each level (must be BEFORE 'reserve' statements)
        int hist_data_recv_counts [L+1];
        MPI_Gather ( &hist_data_chunk, 1, MPI_INT, &hist_data_recv_counts[0], 1, MPI_INT, 0, comm_levels );
        // reserve space on the root
        if (multi_main) HIST_DATA[l] .resize( hist_data_total );
        // === not-so-straighforward gathering with _unequal_ chunks of data
        // strides in the receive buffer for data from each level
        int hist_data_recv_strides [L+1];
        hist_data_recv_strides[0] = 0;
        if (multi_main)
          for (int l_=1; l_<=L; l_++)
            hist_data_recv_strides[l_] = hist_data_recv_strides[l_-1] + hist_data_recv_counts[l_-1];
        // setup send buffer
        mpi_send_buf = multi_main ? MPI_IN_PLACE : HIST_DATA[l].data();
        // gather data into the specified locations within the root
        MPI_Gatherv ( mpi_send_buf, hist_data_chunk, mpi_REAL,
                      HIST_DATA[l].data(), &hist_data_recv_counts[0], &hist_data_recv_strides[0],  mpi_REAL, 0, comm_levels );
      }
    }
    
    #ifndef QUIET
    if (multi_main) std::cout << " done." << std::endl;
    #endif
    
    halt_timer_mpi_mlmc();
}
#endif

void multi_cleanup () {
  
  #ifndef QUIET
  if (multi_main) std::cout << " :: Cleaning up..." << std::endl;
  #endif
  
  // garbage collection from multi_level_init()
  for (int d = 0; d < 2*NDIR; d++) {
    delete [] mpi_rbuf[d];
    delete [] mpi_sbuf[d];
  }
  
  #ifndef QUIET
  if (multi_main) std::cout << "  : -> done." << std::endl;
  #endif
}

void multi_finalise()
{ 
  // make sure all I/O operations are finished
  MPI_Barrier(MPI_COMM_WORLD);
  
  MPI_Finalize();
}
