
#ifndef NUMTYPES_H
#define NUMTYPES_H

#include <math.h>

#include "config.h"

static const int GX = GHOST_SIZE*(NX>1?1:0);
static const int GY = GHOST_SIZE*(NY>1?1:0);
static const int GZ = GHOST_SIZE*(NZ>1?1:0);

// define this for volume index checks.  NOT IMPLEMENTED
// #define INDEX_CHECKING

// dimension related constants

static const int X = 0, Y = 1, Z = 2, NDIR = 3;
static const int XU = 3, YU = 4, ZU = 5;
static const int DX[] = { 1, 0, 0 };
static const int DY[] = { 0, 1, 0 };
static const int DZ[] = { 0, 0, 1 };
static const int ACTIVE_DIRS = (NX>1) + (NY>1) + (NZ>1);

// dimension related functions

static bool HAVE_DIR (int d) { return ( ( d==X?NX:(d==Y?NY:NZ) ) > 1 ); }

// define long_int so that we can easily compile 'long' arrays for MC samples

#include <climits>

//typedef unsigned int long_int;
//static const int NM_MAX = UINT_MAX;

//typedef unsigned long long_int;
//static const long NM_MAX = ULONG_MAX;

typedef unsigned long long long_int;
static const long long NM_MAX = ULLONG_MAX;

// define real so that we can easily compile single or double precision versions of the code.

typedef double real;

// some useful inline numerical functions.

static inline real sqr  (real n)         {return n*n;}
static inline real sign (real n)         {if (n<=0) return -1; else return 1;}
static inline real min  (real a, real b) {if (a < b) return a; else return b;}
static inline real max  (real a, real b) {if (a > b) return a; else return b;}
static inline real ssqrt(real x)         {return sqrt(max(0, x));}
static inline int  mod  (int a, int b)   {int r = a % b; return r < 0 ? r + b : r;}
static inline void Zero (real &in)       {in = 0.0;}
static inline void Zero (int  &in)       {in = 0;}
static inline void Exp  (real &in)       {in = exp(in);}
static inline void Log  (real &in)       {in = log(in);}
static inline void Sqr  (real &in)       {in = log(in);}

// used for structures
template<class T>
real* realCast (T &v) { return (real*)&v; }

// define pi
static const real pi = 3.1415926535897931;

// generic 3d vector type

template<class T>
class raw_vector {
  
  T v[NDIR];

public:
  
  raw_vector(T x, T y, T z) { v[X] = x; v[Y] = y; v[Z] = z; }
  raw_vector() { }
  
  const T &operator[](int i) const { return v[i]; }
  T &operator[](int i) { return v[i]; }
  
  T sqr() const { return v[X]*v[X] + v[Y]*v[Y] + v[Z]*v[Z]; }
  T norm() const { return fabs(v[X]) + fabs(v[Y]) + fabs(v[Z]); }
  
  void zero() {
    for (int i=0; i<NDIR; i++) v[i] = 0;
  }
  
  raw_vector<T> abs() {
    raw_vector<T> a;
    for (int i=0; i<NDIR; i++) a.v[i] = fabs(v[i]);
    return a;
  }
  
  void operator+=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] += a.v[i];
  }
  
  void operator-=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] -= a.v[i];
  }
  
  void operator*=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] *= a.v[i];
  }
  
  void operator/=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] /= a.v[i];
  }
  
  void operator*=(real n)
  {
    for (int i=0; i<NDIR; i++) v[i] *= n;
  }
  
  void operator/=(real n)
  {
    for (int i=0; i<NDIR; i++) v[i] /= n;
  }
  
  // needed for COMPUTE_KURTOSIS=1
  void operator-=(real n)
  {
    for (int i=0; i<NDIR; i++) v[i] -= n;
  }
  // end
  
  friend raw_vector<T> operator+(const raw_vector<T> &a, const raw_vector<T> &b)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i] + b.v[i];
    return r;
  }
  
  friend raw_vector<T> operator-(const raw_vector<T> &a, const raw_vector<T> &b)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i] - b.v[i];
    return r;
  }
  
  friend raw_vector<T> operator*(const raw_vector<T> &a, real n)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i]*n;
    return r;
  }
  
  friend raw_vector<T> operator*(real n, const raw_vector<T> &a)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i]*n;
    return r;
  }
  
  friend raw_vector<T> operator/(const raw_vector<T> &a, real n)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i]/n;
    return r;
  }
  
  // dot product
  friend real operator*(const raw_vector<T> &a, const raw_vector<T> &b)
  { real r=0;
    for (int i=0; i<NDIR; i++) r += a.v[i] * b.v[i];
    return r;
  }
};

// vector of reals
typedef raw_vector<real> vector;

#endif
