
#include "equation.h"
#include "solver-util-hll.h"
#include "solver.h"

///::title:: HLL3R solver (Relaxed 3 wave) | for MHD
///::depends:: solver-mhd-hll-util
///::consts:: HLL_dt=1

void solver_calc_flux(ConservedVars &F, vector &us, vector &Bs,
		      const AllVariables &lV, const AllVariables &rV,
		      int i, int j, int k, int d)
{
  int d1 = (d+1)%NDIR;
  int d2 = (d+2)%NDIR;
  vector pil = -lV.U.B[d]*lV.U.B;
  pil[d] += lV.XU.p + 0.5*lV.U.B.sqr();
  vector pir = -rV.U.B[d]*rV.U.B;
  pir[d] += rV.XU.p + 0.5*rV.U.B.sqr();

  real alpha = (GAMMA + 1)/2;

  real Xl = max(0, lV.XU.u[d] - rV.XU.u[d])
    + max(0, pir[d] - pil[d]) / (lV.U.rho * lV.XU.a + rV.U.rho * rV.XU.cf[d]);
  real Xr = max(0, lV.XU.u[d] - rV.XU.u[d])
    + max(0, pil[d] - pir[d]) / (rV.U.rho * rV.XU.a + lV.U.rho * lV.XU.cf[d]);
  
  real x_l = 1 - Xl/(lV.XU.cf[d] + alpha * Xl);
  real x_r = 1 - Xr/(rV.XU.cf[d] + alpha * Xr);
  
  real a = lV.U.B.sqr()/lV.U.rho + sqr(lV.XU.a)*x_l;
  real a0l = sqrt(max(0,a + sqrt(max(0,sqr(a) - 4*x_l*sqr(lV.XU.a*lV.U.B[d])
				     /lV.U.rho))/(2*x_l)));
  a = rV.U.B.sqr()/rV.U.rho + sqr(rV.XU.a)*x_r;
  real a0r = sqrt(max(0,a + sqrt(max(0,sqr(a) - 4*x_r*sqr(rV.XU.a*rV.U.B[d])
				     /rV.U.rho))/(2*x_r)));
  real cl = lV.U.rho * a0l + alpha * lV.U.rho * Xl;
  real cr = rV.U.rho * a0r + alpha * rV.U.rho * Xr;
  us = (cl * lV.XU.u + cr * rV.XU.u + pil - pir)/(cl + cr);

  if (lV.XU.u[d] - cl/lV.U.rho > 0) {
    F.rho = lV.U.m[d];
    F.m = lV.U.m * lV.XU.u[d] + pil;
    F.E = lV.U.E * lV.XU.u[d] + pil * lV.XU.u;
    F.B = lV.U.B * lV.XU.u[d] - lV.U.B[d] * lV.XU.u;
    Bs = lV.U.B;
    return;
  }
  if (lV.XU.u[d] + cl/lV.U.rho < 0) {
    F.rho = rV.U.m[d];
    F.m = rV.U.m * rV.XU.u[d] + pir;
    F.E = rV.U.E * rV.XU.u[d] + pir * rV.XU.u;
    F.B = rV.U.B * rV.XU.u[d] - rV.U.B[d] * rV.XU.u;
    Bs = rV.U.B;
    return;
  }
  vector pis = (cr*pil + cl*pir - cr*cl*(rV.XU.u - lV.XU.u)) / (cl + cr);
  real rhos, es;
  if (us[d] > 0) {
    rhos = 1/(1/lV.U.rho + (pil[d]-pis[d])/sqr(cl));
    Bs = rhos * (lV.U.B/lV.U.rho + lV.U.B[d]*(pil - pis)/sqr(cl));
    es = lV.XU.p/(lV.U.rho*(GAMMA-1)) +
      0.5*(lV.U.B.sqr()/lV.U.rho - Bs.sqr()/rhos) +
      0.5*(pis.sqr() - pil.sqr())/sqr(cl);
  } else {
    rhos = 1/(1/rV.U.rho + (pir[d]-pis[d])/sqr(cr));
    Bs = rhos * (rV.U.B/rV.U.rho + rV.U.B[d]*(pir - pis)/sqr(cr));
    es = rV.XU.p/(rV.U.rho*(GAMMA-1)) +
      0.5*(rV.U.B.sqr()/rV.U.rho - Bs.sqr()/rhos) +
      0.5*(pis.sqr() - pir.sqr())/sqr(cr);
  }
  es = es*rhos + rhos*us.sqr()/2 + Bs.sqr()/2;
  F.rho = rhos * us[d];
  F.m = rhos * us * us[d] + pis;
  F.E = es * us[d] + pis * us;
  F.B = Bs * us[d] - Bs[d] * us;
}
