
#include "equation.h"
#include "solver-util-ec.h"

void ec_flux (ConservedVars &ecF, const ConservedVars &lU, const ConservedVars &rU, const XtraVars &lXU, const XtraVars &rXU, int d)
{
  // caution: this is NOT the same (though looks similar in parts) as calc_point_flux(ecF, avg(lU, rU), d)
  ecF.h       = avg ( lU.h, rU.h ) * avg ( lXU.u[d], rXU.u[d] );
  ecF.hu[d]   = avg ( lU.h, rU.h ) * sqr ( avg ( lXU.u[d], rXU.u[d] ) ) + GRAVITY * avg ( sqr(lU.h), sqr(rU.h) ) / 2.0;
  ecF.hu[1-d] = avg ( lU.h, rU.h ) * avg ( lXU.u[d], rXU.u[d] ) * avg ( lXU.u[1-d], rXU.u[1-d] );
}

// 2nd order accurate energy conservative (EC) source quadrature
void compute_source (const int &i, const int &j) {
  Source[i][j][0].hu[X] -= avg ( U[i]  [j][0].h, U[i+1][j][0].h ) * jump( Xtra[i]  [j][0].b, Xtra[i+1][j][0].b );
  Source[i][j][0].hu[X] -= avg ( U[i-1][j][0].h, U[i]  [j][0].h ) * jump( Xtra[i-1][j][0].b, Xtra[i]  [j][0].b );
  Source[i][j][0].hu[X] *= GRAVITY / (2*dx[X]);
  if (HAVE_DIR(Y))
    Source[i][j][0].hu[Y] -= avg ( U[i][j]  [0].h, U[i][j+1][0].h ) * jump( Xtra[i][j]  [0].b, Xtra[i][j+1][0].b );
    Source[i][j][0].hu[Y] -= avg ( U[i][j-1][0].h, U[i][j]  [0].h ) * jump( Xtra[i][j-1][0].b, Xtra[i][j]  [0].b );
    Source[i][j][0].hu[Y] *= GRAVITY / (2*dx[Y]);
}

real avg  (const real &left, const real &right)
{
  return (right + left) / 2.0;
}

real jump (const real &left, const real &right)
{
  return right - left;
}

void compute_R (real R [3][3], const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d)
{
  real sqrt_gh = sqrt ( GRAVITY * avg(lCV.h, rCV.h) );
  
  // 1st column of R
  R[0][0] = 1;
  R[1+d][0] = avg(lXV.u[d], rXV.u[d]) - sqrt_gh;
  R[2-d][0] = avg(lXV.u[1-d], rXV.u[1-d]);
  
  // 2nd column of R
  R[0][1] = 0;
  R[1+d][1] = 0;
  if (d == X)
    R[2][1] = sqrt_gh;
  else // (d == Y)
    R[1][1] = -sqrt_gh;
  
  // 3rd column of R
  R[0][2] = 1;
  R[1+d][2] = avg(lXV.u[d], rXV.u[d]) + sqrt_gh;
  R[2-d][2] = avg(lXV.u[1-d], rXV.u[1-d]);
  
  real scaling = sqrt(2*GRAVITY);
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      R[a][b] /= scaling;
}

void compute_Rinv (real R [3][3], const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d)
{
  real sqrt_gh = sqrt ( GRAVITY * avg(lCV.h, rCV.h) );
  
  real a = avg(lXV.u[d], rXV.u[d]);
  real b = sqrt_gh;
  real c = avg(lXV.u[1-d], rXV.u[1-d]);
  
  // 1st column of R inverse
  R[0][0] = (a*b + b*b) / (2*b*b);
  R[1+d][0] = -0.5*b;
  R[2-d][0] = 0;
  
  // 2nd column of R inverse
  if (d == X) {
    R[0][1] = -c/b;
    R[1][1] = 0;
    R[2][1] = 1/b;
  }
  else { // (d == Y)
    R[0][1] = c/b;
    R[1][1] = -1/b;
    R[2][1] = 0;
  }
  
  // 3rd column of R inverse
  R[0][2] = (a*b + b*b) / (2*b*b);
  R[1+d][2] = 0.5*b;
  R[2-d][2] = 0;
  
  real scaling = sqrt(2*GRAVITY);
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      R[a][b] /= scaling;
}

void transpose (real R [3][3])
{
  real tmp;
  for (int a=1; a<3; a++)
    for (int b=0; b<a; b++) {
      tmp = R[a][b];
      R[a][b] = R[b][a];
      R[b][a] = tmp;
    }
}
