
#ifndef VOLUMES_H
#define VOLUMES_H

#include <math.h>
#include <malloc.h>
#include <iostream>
#include <cstring>

#include "numtypes.h"
#include "structures.h"

// generic 3d volume array type (dynamic version)
// supports += and -= operators for hierarchical grids

template<class T, int GHOST, int VX0, int VY0, int VZ0>
class raw_volume {

private:
  
  int getLoc (int i, int j, int k) { return (i-VX0)*VNY*VNZ + (j-VY0)*VNZ + (k-VZ0); }
  
public:
  
  // pointer to actual array
  T* data;
  
  // grid sizes
  int NXl;
  int NYl;
  int NZl;
  
  // grid sizes including ghost cells
  int VNX;
  int VNY;
  int VNZ;
  
  void zero () {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++)
          Zero(data[getLoc(i,j,k)]);
  }
  
  // zero length allocations prevention
  raw_volume () : NXl(1), NYl(1), NZl(1), VNX(1 + 2*GX*GHOST), VNY(1 + 2*GY*GHOST), VNZ(1 + 2*GZ*GHOST) {
    data = (T*) malloc ( VNX * VNY * VNZ * sizeof(T) );
    if ( data == NULL ) {
      std::cout << " :: ERROR: out of memory for volume allocation." << std::endl;
      throw 1;
    }
    zero ();
  }
  
  raw_volume (int NXl_, int NYl_, int NZl_) : NXl(NXl_), NYl(NYl_), NZl(NZl_), VNX(NXl_ + 2*GX*GHOST), VNY(NYl_ + 2*GY*GHOST), VNZ(NZl_ + 2*GZ*GHOST) {
    data = (T*) malloc ( VNX * VNY * VNZ * sizeof(T) );
    if ( data == NULL ) {
      std::cout << " :: ERROR: out of memory for volume allocation." << std::endl;
      throw 1;
    }
    zero();
  }
  
  raw_volume (const raw_volume &a) {
    NXl = a.NXl;
    NYl = a.NYl;
    NZl = a.NZl;
    VNX = a.VNX;
    VNY = a.VNY;
    VNZ = a.VNZ;
    data = (T*) malloc ( VNX * VNY * VNZ * sizeof(T) );
    if ( data == NULL ) {
      std::cout << " :: ERROR: out of memory for volume allocation." << std::endl;
      throw 1;
    }
    memcpy ( data, a.data, VNX * VNY * VNZ * sizeof(T) );
  }
  
  void reinit (int NXl_, int NYl_, int NZl_) {
    
    if ( NXl == NXl_ && NYl == NYl_ && NZl == NZl_ ) {
      zero();
      return;
    }
    
    NXl = NXl_;
    NYl = NYl_;
    NZl = NZl_;
    VNX = NXl_ + 2*GX*GHOST;
    VNY = NYl_ + 2*GY*GHOST;
    VNZ = NZl_ + 2*GZ*GHOST;
    T* data_ = (T*) realloc ( data, VNX * VNY * VNZ * sizeof(T) );
    if ( data_ == NULL ) {
      free (data);      // memory leak prevention
      std::cout << " :: ERROR: out of memory for volume reinitialization." << std::endl;
      throw 1;
    }
    data = data_;
    zero();
  }
  
  void reset () { zero(); }
  
  ~raw_volume () {
    free (data);        // memory leak prevention
  }
  
  // support for subscript operator [][][]
  struct slice_X {
    T* d;
    int VNZ;
    slice_X ( T *d_, int shift, int VNY_, int VNZ_ ) : d(&d_[shift*VNY_*VNZ_]), VNZ(VNZ_) {}
    
    struct slice_Y {
      T* d;
      slice_Y ( T *d_, int shift, int VNZ_ ) : d(&d_[shift*VNZ_]) {}
      
      T& operator[](const int k) const { return d[k]; }
    };
    
    slice_Y operator[](const int j) const { return slice_Y (d, j-VY0, VNZ); }
  };
  slice_X operator[](const int i) const { return slice_X (data, i-VX0, VNY, VNZ); }
  
  // validates compatibility of volumes for use with operators
  void validate (const raw_volume &a) const {
    if ( NXl != a.NXl || NYl != a.NYl || NZl != a.NZl ) {
      std::cout << " :: ERROR: non-compatible volume sizes." << std::endl;
      throw 1;
    }
  }
  
  void exp() {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++)
          Exp(data[getLoc(i,j,k)]);
  }
  
  void log() {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++)
          Log(data[getLoc(i,j,k)]);
  }
  
  void nonnegative() {
    for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++)
            NonNegative(data[getLoc(i,j,k)]);
  }
  
  void sqr () {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++) {
          data[getLoc(i,j,k)] = data[getLoc(i,j,k)] * data[getLoc(i,j,k)];
    }
  }
  
  void operator+=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] += a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] += a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }

  void operator-=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] -= a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] -= a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }
  
  void operator*=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] *= a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] *= a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }

  void operator/=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] /= a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] /= a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }
  
  void operator+=(const real r)
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
	for (int k=VZ0; k<VNZ+VZ0; k++) {
	  data[getLoc(i,j,k)] += r;
	}
  }
  
  void operator-=(const real r)
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
	for (int k=VZ0; k<VNZ+VZ0; k++) {
	  data[getLoc(i,j,k)] -= r;
	}
  }
  
  void operator*=(const real r)
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
	for (int k=VZ0; k<VNZ+VZ0; k++) {
	  data[getLoc(i,j,k)] *= r;
	}
  }
  
  void operator/=(const real r)
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
	for (int k=VZ0; k<VNZ+VZ0; k++) {
	  data[getLoc(i,j,k)] /= r;
	}
  }
  
  // === !!! operators creating NEW objects are very risky and can lead to memory inefficiency !!! be careful :)
  
  // needed for variance computation during parallel runs
  friend raw_volume operator-(const raw_volume &a, const raw_volume &b)
  {
    a.validate(b);
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] - b[i][j][k];
        }
    return v;
  }
  
  // needed for numerically stable variance computation
  friend raw_volume operator/(const raw_volume &a, const real r)
  {
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] / r;
        }
    return v;
  }
  
  // needed for COMPUTE_KURTOSIS=1
  
  friend raw_volume operator*(const raw_volume &a, const raw_volume &b)
  {
    a.validate(b);
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] * b[i][j][k];
        }
    return v;
  }
  
  friend raw_volume operator*(const raw_volume &a, const real r)
  {
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] * r;
        }
    return v;
  }
  
  // I/O routines
  
  void print ()
  {
    if (ACTIVE_DIRS == 3) {
      std::cout << "print() is not implemented for 3d" << std::endl;
      return;
    }
    
    for (int i=VX0; i<VNX+VX0; i++) {
      for (int j=VY0; j<VNY+VY0; j++)
        std::cout << data[getLoc(i,j,0)] << " ";
      std::cout << std::endl;
    }
  }
};

// 3d volume array type including ghost zones

template<class T, int GHOST>
class volume: public raw_volume<T, GHOST, -GX*GHOST, -GY*GHOST, -GZ*GHOST> {
public:
  volume () : raw_volume<T, GHOST, -GX*GHOST, -GY*GHOST, -GZ*GHOST> () {}
  volume (int NX_, int NY_, int NZ_) : raw_volume<T, GHOST, -GX*GHOST, -GY*GHOST, -GZ*GHOST> (NX_, NY_, NZ_) {}
};

// === various distance functions

// evaluates the distances between two points
static real dist_euclidean (const vector &x1, const vector &x2) {
  
  real dist = fabs(x2[X]-x1[X]);
  
  if (!HAVE_DIR(Y))
    return dist;
  
  else {
    dist = sqr(dist);
    dist += sqr(x2[Y]-x1[Y]);
    if (HAVE_DIR(Z))
      dist += sqr(x2[Z]-x1[Z]);
    return sqrt(dist);
  }
}

// evaluates the distances between two points in a periodic domain
static real dist_periodic (const vector &x1, const vector &x2) {
  
  real dist = min( fabs(x2[X]-x1[X]), MAXX - fabs(x2[X]-x1[X]) );
  
  if (!HAVE_DIR(Y))
    return dist;
  
  else {
    dist = sqr(dist);
    dist += sqr( min( fabs(x2[Y]-x1[Y]), MAXY - fabs(x2[Y]-x1[Y]) ) );
    if (HAVE_DIR(Z))
      dist += sqr( min( fabs(x2[Z]-x1[Z]), MAXZ - fabs(x2[Z]-x1[Z]) ) );
    return sqrt(dist);
  }
}

#endif
